# Author: Nathan Trouvain at 17/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from reservoirpy.nodes import RLS


def test_rls_init():
    node = RLS(10, output_dim=3)

    data = np.ones((100,))
    # y = np.ones((1, 3))
    res = node(data)

    assert node.Wout.shape == (100, 3)
    assert node.bias.shape == (3,)
    assert node.alpha == 10

    data = np.ones((10000, 100))
    res = node.run(data)
    assert res.shape == (10000, 3)

    # with initialized Wout and bias
    Wout = np.ones((100, 3))
    bias = np.ones((3,))
    data = np.ones((100,))
    # unspecified dimensions
    node = RLS(10, Wout=Wout, bias=bias)
    res = node(data)
    # correct specified dimensions
    node = RLS(10, Wout=Wout, bias=bias, input_dim=100, output_dim=3)
    res = node(data)
    # incorrect specified dimensions
    with pytest.raises(ValueError):
        node = RLS(10, Wout=Wout, bias=bias, input_dim=101, output_dim=1)
    with pytest.raises(ValueError):
        bias = np.ones((10,))
        node = RLS(bias=bias, output_dim=1)


def test_rls_train_one_step():
    node = RLS(10)

    x = np.ones((10, 5, 2))
    y = np.ones((10, 5, 10))

    for x, y in zip(x, y):
        res = node.partial_fit(x, y)

    assert node.Wout.shape == (2, 10)
    assert node.bias.shape == (10,)
    assert node.alpha == 10

    data = np.ones((10000, 2))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_rls_train():
    node = RLS(alpha=1e-6)

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.partial_fit(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.005, decimal=4)
    assert node.bias.shape == (10,)
    assert_array_almost_equal(node.bias, np.ones((10,)) * 0.5, decimal=4)

    node = RLS(1e-6, fit_bias=False)

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.partial_fit(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (10,)
    assert_array_almost_equal(node.bias, np.ones((10,)) * 0.0, decimal=4)

    node = RLS(1e-6, fit_bias=True)

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = node.partial_fit(x, y)

    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.005, decimal=4)
    assert node.bias.shape == (10,)
    assert_array_almost_equal(node.bias, np.ones((10,)) * 0.5, decimal=4)

    data = np.ones((1000, 100))
    res = node.run(data)

    assert res.shape == (1000, 10)



"""Tools for CPU-parallelized hyperparameter optimization."""

import json
import os
import time
import warnings
from glob import glob

import numpy as np

from reservoirpy.utils.random import rand_generator


def _get_conf_from_json(confpath):
    if not (os.path.isfile(confpath)):
        raise FileNotFoundError(f"Training conf '{confpath}' not found.")
    else:
        config = {}
        with open(confpath, "r") as f:
            config = json.load(f)
        return _parse_config(config)


def _parse_config(config):
    required_args = ["exp", "hp_max_evals", "hp_method", "hp_space"]
    for arg in required_args:
        if config.get(arg) is None:
            raise ValueError(f"No {arg} argument found in configuration file.")

    if config["hp_method"] not in ["random"]:
        raise ValueError(
            f"Unknown algorithm: {config['hp_method']}. "
            "Available algorithms for CPU parallelized research: 'random'."
        )

    config["hp_space"] = {arg: _parse_searchspace(specs) for arg, specs in config["hp_space"].items()}

    return config


def _parse_searchspace(specs):
    if specs[0] == "choice":
        # specs: ['choice', e0, ..., eN]
        return lambda rng: specs[1:][rng.integers(0, len(specs) - 1)]
    if specs[0] == "randint":
        # specs: ['randint', low, high]
        return lambda rng: rng.integers(specs[1], specs[2])
    if specs[0] == "uniform":
        # specs: ['uniform', low, high]
        return lambda rng: rng.uniform(specs[1], specs[2])
    if specs[0] == "quniform":
        # specs: ['quniform', low, high, q]
        return lambda rng: np.round(rng.uniform(specs[1], specs[2]) / specs[3]) * specs[3]
    if specs[0] == "loguniform":
        # specs: ['loguniform', low, high]
        return lambda rng: np.exp(rng.uniform(np.log(specs[1]), np.log(specs[2])))
    if specs[0] == "qloguniform":
        # specs: ['qloguniform', low, high, q]
        return lambda rng: np.round(np.exp(rng.uniform(np.log(specs[1]), np.log(specs[2]))) / specs[3]) * specs[3]
    if specs[0] == "normal":
        # specs: ['normal', mu, sigma]
        return lambda rng: specs[1] + specs[2] * rng.randn()
    if specs[0] == "qnormal":
        # specs: ['qnormal', mu, sigma, q]
        return lambda rng: np.round((specs[1] + specs[2] * rng.randn()) / specs[3]) * specs[3]
    if specs[0] == "lognormal":
        # specs: ['lognormal', mu, sigma]
        return lambda rng: np.exp(np.log(specs[1]) + np.log(specs[2]) * rng.randn())
    if specs[0] == "qlognormal":
        # specs: ['qlognormal', mu, sigma, q]
        return lambda rng: np.round(np.exp(np.log(specs[1]) + np.log(specs[2]) * rng.randn()) / specs[3]) * specs[3]


def _worker(objective, dataset, config, kwargs):
    """
    Worker function to evaluate the objective with given parameters.
    Measures execution time and handles exceptions.
    Returns the parameters, result dictionary, and filename for saving results.
    """
    try:
        start = time.time()

        returned_dict = objective(dataset, config, **kwargs)

        end = time.time()
        duration = end - start

        returned_dict["status"] = "ok"
        returned_dict["start_time"] = start
        returned_dict["duration"] = duration

        save_file = f"{returned_dict['loss']:.7f}_results"

    except Exception as e:
        raise e
        start = time.time()

        returned_dict = {
            "status": "fail",
            "start_time": start,
            "error": str(e),
        }

        save_file = f"ERR{start}_results"

    return kwargs, returned_dict, save_file


def _get_report_path(exp_name, base_path=None):
    base_path = "." if base_path is None else base_path

    report_path = os.path.join(base_path, exp_name, "results")

    if not (os.path.isdir(base_path)):
        os.mkdir(base_path)

    if not (os.path.isdir(os.path.join(base_path, exp_name))):
        os.mkdir(os.path.join(base_path, exp_name))

    if not (os.path.isdir(report_path)):
        os.mkdir(report_path)

    return report_path


def parallel_research(
    objective,
    dataset,
    config_path,
    report_path=None,
    n_jobs=-1,
    inner_max_num_threads=None,
):
    """
    Executes a parallelized hyperparameter search on the provided
    objective function using the configuration specified.
    This function uses **process-based parallelism** through the `loky` backend of `joblib`.

    For more details on hyper-parameter search using ReservoirPy, take a look at
    :ref:`/user_guide/hyper.ipynb`.

    Parameters
    ----------
    objective : Callable
        Objective function defining the function to
        optimize. Must be able to receive the dataset argument and
        all parameters sampled by hyperopt during the search. These
        parameters must be keyword arguments only without default value
        (this can be achieved by separating them from the other arguments
        with an empty starred expression. See examples for more info.)
    dataset : tuple or lists or arrays of data
        Argument used to pass data to the objective function during
        the hyperopt run. It will be passed as is to the objective
        function : it can be in whatever format.
    config_path : str or Path
        Path to the hyperopt experimentation configuration file used to
        define this run.
    report_path : str, optional
        Path to the directory where to store the results of the run. By default,
        this directory is set to be {name of the experiment}/results/.
    n_jobs: int, default to -1
        Maximum number of concurrent processes to run.
        If set to -1, it is set to the number of CPUs.
    inner_max_num_threads: int, optional
        Maximum number of threads allowed within each parallel job (e.g., used by
        libraries like NumPy). If set to a positive number, it will limit the
        number of threads per process. If set to -1, there will be no restriction.
        By default, it is calculated as `max(1, number of trials // number of jobs)`.
    """
    from joblib import Parallel, cpu_count, delayed, parallel_config
    from tqdm import tqdm

    config = _get_conf_from_json(config_path)
    report_path = _get_report_path(config["exp"], report_path)

    search_space = config["hp_space"]
    max_evals = config["hp_max_evals"]

    if inner_max_num_threads is None:
        inner_max_num_threads = max(1, cpu_count() // max_evals)
    elif inner_max_num_threads == -1:
        inner_max_num_threads = None

    rng = rand_generator(config.get("seed"))

    params_list = [{arg: f(rng) for arg, f in search_space.items()} for _ in range(max_evals)]

    best_params = None
    best_loss = float("inf")

    with parallel_config(backend="loky", inner_max_num_threads=inner_max_num_threads):
        it = Parallel(n_jobs=n_jobs, return_as="generator_unordered")(
            delayed(_worker)(objective, dataset, config, kwargs) for kwargs in params_list
        )

    pbar = tqdm(iterable=it, total=len(params_list), unit="trial", postfix="best loss=?")

    for kwargs, returned_dict, save_file in pbar:
        try:
            for key in kwargs:
                if isinstance(kwargs[key], np.integer):
                    kwargs[key] = int(kwargs[key])
            json_dict = {"returned_dict": returned_dict, "current_params": kwargs}
            save_file = os.path.join(report_path, save_file)
            nb_save_file_with_same_loss = len(glob(f"{save_file}*"))
            save_file = f"{save_file}_{nb_save_file_with_same_loss+1}call.json"
            with open(save_file, "w+") as f:
                json.dump(json_dict, f, indent=2)
        except Exception as e:
            warnings.warn("Results of current simulation were NOT saved " "correctly to JSON file.")
            warnings.warn(str(e))

        if returned_dict["loss"] < best_loss:
            best_params = kwargs
            best_loss = returned_dict["loss"]
            pbar.set_postfix({"best loss": best_loss})

    return best_params, best_loss



"""
====================================
Node API (:mod:`reservoirpy.node`)
====================================

Note
----

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`
- **Learn how to subclass Node to make your own**: :ref:`create_new_node`


**Simple tools for complex reservoir computing architectures.**

The Node API features a simple implementation of computational graphs, similar
to what can be found in other popular deep learning and differentiable calculus
libraries. It is however simplified and made the most flexible possible by
discarding the useless "fully differentiable operations" functionalities. If
you wish to use learning rules making use of chain rule and full
differentiability of all operators, ReservoirPy may not be the tool you need
(actually, the whole paradigm of reservoir computing might arguably not be the
tool you need).

The Node API is composed of a base :py:class:`Node` class that can be described
as a stateful recurrent operator able to manipulate streams of data. A
:py:class:`Node` applies a `_step` function on some data, and then stores the
result in its `state` attribute. The `_step` operation can be a function
depending on the data, on the current `state` vector of the Node and on the Node
parameters.

Nodes can also be connected together to form a :py:class:`Model`. Models hold
references to the connected nodes and make data flow from one node to
the next, allowing to create *deep* models and other more complex
architectures and computational graphs.
:py:class:`Model` is essentially a subclass of :py:class:`Node`,
that can also be connected to other nodes and models.

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`
- **Learn how to subclass Node to make your own**: :ref:`create_new_node`

.. currentmodule:: reservoirpy.node


.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

    Node
    TrainableNode
    OnlineNode
    ParallelNode
"""

from abc import ABC, abstractmethod
from itertools import repeat
from typing import Iterable, Optional, Sequence, Union

import numpy as np
from joblib import Parallel, delayed

from reservoirpy.utils.data_validation import check_node_input, check_timestep

from .type import NodeInput, State, Timeseries, Timestep, is_multiseries


class Node(ABC):
    """Generic Node ABC

    All Nodes should inherit this class.

    """

    #: True if the Node has been initialized
    initialized: bool
    #: Expected dimension of the Node input. Can be None before initialization
    input_dim: int = None
    #: Expected dimension of the Node input. Can be None before initialization
    output_dim: int = None
    #: Current state of the Node. Must have "out" as one of the keys.
    state: State
    #: Optional name of the Node.
    name: Optional[str] = None

    @abstractmethod
    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        """Define input and output dimensions, and instantiate variables.

        Only called once, before fitting or running the node.

        Parameters
        ----------
        x : array of shape (input_dim,) or (timestep, input_dim)
            Input data to the node.
        y : None
            Training data to the node. As it is not a trainable node, ``y`` is
            expected to be ``None``.
        """
        ...  # pragma: no cover

    @abstractmethod
    def _step(self, state: State, x: Timestep) -> State:
        ...  # pragma: no cover

    def step(self, x: Optional[Timestep] = None) -> Timestep:
        """Call the Node function on a single step of data and update
        the state of the Node.

        Parameters
        ----------
        x : array of shape (input_dim,), optional
            One single step of input data. If None, an empty array is used
            instead and the Node is assumed to have an input_dim of 0

        Returns
        -------
        array of shape (output_dim,)
            An output vector.
        """
        # TODO: stateful argument (for every step, run, fit, train, ...)
        # Auto-regressive mode
        if x is None:
            x: Timestep = np.empty((0,))
        check_timestep(x, expected_dim=self.input_dim)

        if not self.initialized:
            self.initialize(x)

        new_state = self._step(self.state, x)

        self.state = new_state
        return new_state["out"]

    def _run(self, state: State, x: Timeseries) -> tuple[State, Timeseries]:
        current_state = state
        n_timesteps = x.shape[-2]

        output = np.empty((n_timesteps, self.output_dim))
        for i, x_step in enumerate(x):
            current_state = self._step(state=current_state, x=x_step)
            output[i] = current_state["out"]

        return current_state, output

    def run(self, x: Optional[NodeInput] = None, iters: Optional[int] = None, workers=1) -> NodeInput:
        """Run the Node on a sequence of data.
        Can update the state of the
        Node several times.

        Parameters
        ----------
        x : array-like of shape ([n_inputs,] timesteps, input_dim) or list of \
                arrays of shape (timesteps, input_dim), optional
            A timeseries, array of shape (timesteps, features), or a sequence of
            timeseries. Input of the Node.
        iters : int, optional
            If ``x`` is ``None``, a dimensionless timeseries of length ``iters``
            is used instead.
        workers : int, default to 1
            Number of workers used for parallelization. If set to -1, all available
            workers (threads or processes) are used.

        Returns
        -------
        array of shape ([n_inputs,] timesteps, output_dim) or list of arrays
            A sequence of output vectors.
        """
        # Auto-regressive mode
        if x is None:
            x = np.empty((iters, 0))
        check_node_input(x, expected_dim=self.input_dim)

        if not self.initialized:
            self.initialize(x)

        initial_state = self.state

        if is_multiseries(x):
            output = Parallel(n_jobs=workers)(delayed(self._run)(initial_state, timeseries) for timeseries in x)
            states, result = zip(*output)
            final_state = states[-1]

            if isinstance(x, np.ndarray):
                result = np.array(result)
            else:
                result = list(result)
        else:
            final_state, result = self._run(initial_state, x)

        self.state = final_state
        return result

    def predict(self, x: Optional[NodeInput] = None, iters: Optional[int] = None, workers=1) -> NodeInput:
        """Alias for :py:meth:`~.Node.run`

        Run the Node on a sequence of data.
        Can update the state of the
        Node several times.

        Parameters
        ----------
        x : array-like of shape ([n_inputs,] timesteps, input_dim) or list of
                arrays of shape (timesteps, input_dim), optional
            A sequence of data of shape (timesteps, features).
        iters : int, optional
            If ``x`` is ``None``, a dimensionless timeseries of length ``iters``
            is used instead.

        Returns
        -------
        array of shape ([n_inputs,] timesteps, output_dim) or list of arrays
            A sequence of output vectors.
        """
        return self.run(x=x, iters=iters, workers=workers)

    def reset(self) -> State:
        """Reset all Node state

        Returns
        -------
        dict[str, np.array]: previous state of the Node.
        """
        previous_state = self.state
        self.state = {key: np.zeros(val.shape) for key, val in self.state.items()}
        return previous_state

    def _set_input_dim(self, x: Optional[Union[NodeInput, Timestep]]):
        if x is None:
            return
        if isinstance(x, Sequence):
            if len(x) == 0:
                return
            input_dim = x[0].shape[-1]
        else:
            input_dim = x.shape[-1]  # works for both timesteps & timeseries
        if self.input_dim is not None and self.input_dim != input_dim:
            raise ValueError(
                f"Trying to set {self} input_dim to {input_dim} while it has already been set to {self.input_dim}"
            )
        self.input_dim = input_dim

    def _set_output_dim(self, y: Optional[Union[NodeInput, Timestep]]):
        if y is None:
            return
        if isinstance(y, Sequence):
            if len(y) == 0:
                return
            output_dim = y[0].shape[-1]
        else:
            output_dim = y.shape[-1]
        if self.output_dim is not None and self.output_dim != output_dim:
            raise ValueError(
                f"Trying to set {self} input_dim to {output_dim} while it has already been set to {self.output_dim}"
            )
        self.output_dim = output_dim

    def __call__(self, x: Optional[Timestep] = None) -> Timestep:
        return self.step(x)

    def __repr__(self):
        if self.name is not None:
            return self.name
        else:
            return self.__class__.__name__

    def __rshift__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import link

        return link(self, other)

    def __rrshift__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import link

        return link(other, self)

    def __lshift__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=other, receiver=self)

    def __rlshift__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=self, receiver=other)

    def __and__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import merge

        return merge(self, other)

    def __rand__(self, other: Union["Node", "Model", Sequence[Union["Node", "Model"]]]) -> "Model":
        from .ops import merge

        return merge(other, self)


class TrainableNode(Node):
    """Node that can be trained.

    :py:class:`reservoirpy.node.TrainableNode` implements the stateful method
    :py:meth:`~.TrainableNode.fit`.

    """

    @abstractmethod
    def fit(self, x: NodeInput, y: Optional[NodeInput] = None, warmup: int = 0) -> "TrainableNode":
        """Offline fitting method of a Node.

        Parameters
        ----------
        x : list or array-like of shape ([series, ] timesteps, input_dim), optional
            Input sequences dataset.
        y : list or array-like of shape ([series], timesteps, output_dim), optional
            Teacher signals dataset. If None, the method will try to fit the
            Node in an unsupervised way, if possible.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.

        Returns
        -------
        Node
            Node trained offline.
        """
        ...  # pragma: no cover


class OnlineNode(TrainableNode):
    """Node that can be trained in an online fashion.

    :py:class:`reservoirpy.node.OnlineNode` implements the stateful method
    :py:meth:`~.OnlineNode.partial_fit` and :py:meth:`~.OnlineNode._learning_step`.

    """

    @abstractmethod
    def _learning_step(self, x: Timestep, y: Optional[Timestep]) -> Timestep:
        ...  # pragma: no cover

    @abstractmethod
    def partial_fit(self, x: Timeseries, y: Optional[Timeseries]) -> Timeseries:
        """Fit the Node in an online fashion.

        This method both trains the Node parameters and produce predictions on
        the run. Calling :py:meth:`partial_fit` updates the Node without
        resetting the parameters, unlike :py:meth:`fit`.

        Parameters
        ----------
        x : array-like of shape (timesteps, input_dim)
            Input sequence of data.
        y : array-like of shape (timesteps, output_dim), optional.
            Target sequence of data. If None, the Node will train in an
            unsupervised way, if possible.

        Returns
        -------
        array of shape (timesteps, output_dim)
            All outputs computed during the training.
        """
        ...  # pragma: no cover

    def fit(self, x: NodeInput, y: Optional[NodeInput] = None, warmup: int = 0) -> "OnlineNode":
        # Re-initialize in any case
        self.initialize(x, y)
        check_node_input(x, expected_dim=self.input_dim)
        if y is not None:
            check_node_input(y)

        if is_multiseries(x):
            if y is None:
                y = repeat(None)
            for x_ts, y_ts in zip(x, y):
                _y_pred_current = self.partial_fit(x_ts[warmup:], None if y_ts is None else y_ts[warmup:])
        else:
            _y_pred = self.partial_fit(x[warmup:], None if y is None else y[warmup:])

        return self


class ParallelNode(TrainableNode, ABC):
    """Node that can be trained in parallel.

    :py:class:`reservoirpy.node.ParallelNode` implements the methods
    :py:meth:`~.ParallelNode.master` and :py:meth:`~.ParallelNode.worker`.

    """

    @abstractmethod
    def worker(self, x: Timeseries, y: Optional[Timeseries]):
        ...  # pragma: no cover

    @abstractmethod
    def master(self, generator: Iterable):
        ...  # pragma: no cover

    def fit(
        self,
        x: NodeInput,
        y: Optional[NodeInput] = None,
        warmup: int = 0,
        workers: int = 1,
    ) -> "ParallelNode":
        check_node_input(x, expected_dim=self.input_dim)
        if y is not None:
            check_node_input(y)

        if not self.initialized:
            self.initialize(x, y)

        # Multi-series
        if is_multiseries(x):
            parallel_operator = Parallel(n_jobs=workers, return_as="generator")
            if y is None:
                results = parallel_operator(delayed(self.worker)(x_ts) for x_ts in x)
            else:
                results = parallel_operator(
                    delayed(self.worker)(x_ts[warmup:], y_ts[warmup:]) for x_ts, y_ts in zip(x, y)
                )

        # Single timeseries
        else:
            results = (self.worker(x[warmup:], y[warmup:]) for _ in range(1))

        self.master(results)

        return self



import numpy as np
import pytest
from numpy.testing import assert_allclose

from ...mat_gen import ring
from .. import Ridge
from ..local_plasticity_reservoir import LocalPlasticityReservoir


def test_lsp_init():
    x = np.ones((10, 5))
    res = LocalPlasticityReservoir(100, input_dim=5)

    res.initialize(x)

    assert res.W.shape == (100, 100)
    assert res.Win.shape == (100, 5)

    res = LocalPlasticityReservoir(100)

    out = res.run(x)

    assert out.shape == (10, 100)
    assert res.W.shape == (100, 100)
    assert res.Win.shape == (100, 5)

    with pytest.raises(ValueError):
        _ = LocalPlasticityReservoir(local_rule="oja")


def test_lsp_rules():
    x = np.ones((10, 5))

    res = LocalPlasticityReservoir(100, local_rule="oja")
    _ = res.fit(x)
    res = LocalPlasticityReservoir(100, local_rule="anti-oja")
    _ = res.fit(x)
    res = LocalPlasticityReservoir(100, local_rule="hebbian")
    _ = res.fit(x)
    res = LocalPlasticityReservoir(100, local_rule="anti-hebbian")
    _ = res.fit(x)
    res = LocalPlasticityReservoir(100, local_rule="bcm")
    _ = res.fit(x)

    with pytest.raises(ValueError):
        res = LocalPlasticityReservoir(100, local_rule="anti-bcm")

    with pytest.raises(ValueError):
        res = LocalPlasticityReservoir(100, local_rule="anti_oja")


def test_local_synaptic_plasticity():
    rng = np.random.default_rng(seed=1)
    x = rng.normal(size=(100, 5))
    X = [x[:10], x[:20]]

    res = LocalPlasticityReservoir(100, local_rule="hebbian", epochs=2, seed=0)

    res.fit(x)
    res.fit(X)

    assert res.W.shape == (100, 100)

    res = LocalPlasticityReservoir(
        100, local_rule="oja", epochs=10, eta=1e-3, synapse_normalization=True, seed=10
    )
    res.initialize(x)

    initial_Wvals = res.W.data.copy()

    res.fit(x)
    res.fit(X)

    assert not np.allclose(initial_Wvals, res.W.data)


def test_lsp_matrices():
    rng = np.random.default_rng(seed=2504)
    x = rng.normal(size=(100, 5))

    W = ring(10, 10)

    lspres_ring = LocalPlasticityReservoir(W=W, seed=2504)
    lspres_rand = LocalPlasticityReservoir(units=10, seed=2504)

    lspres_ring.fit(x)
    lspres_rand.fit(x)

    assert not np.allclose(lspres_ring.W.toarray(), lspres_rand.W.toarray())

    # test dense matrix
    W = rng.normal(size=(10, 10))
    res = LocalPlasticityReservoir(W=W)
    res.fit(x)



"""
===================================
Models (:class:`reservoirpy.Model`)
===================================

Note
----

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`


Models are an extension of the Node API. They allow to combine nodes into
complex computational graphs, to create complicated Reservoir Computing
architecture like *Deep Echo State Networks*.

See :ref:`model` to learn more about how to create and manipulate
a :py:class:`Model`.

.. currentmodule:: reservoirpy.model

.. autoclass:: Model

   .. rubric:: Methods

   .. autosummary::

      ~Model.initialize
      ~Model.step
      ~Model.run
      ~Model.predict
      ~Model.fit
      ~Model.partial_fit


   .. rubric:: Attributes

   .. autosummary::

    ~Model.nodes
    ~Model.edges
    ~Model.inputs
    ~Model.outputs
    ~Model.named_nodes
    ~Model.trainable_nodes
    ~Model.execution_order
    ~Model.parents
    ~Model.children
    ~Model.is_trainable
    ~Model.is_multi_input
    ~Model.is_multi_output
    ~Model.is_parallel
    ~Model.initialized


"""

# Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict
from typing import Any, Mapping, Optional, Sequence, Union

import numpy as np
from joblib import Parallel, delayed

from reservoirpy.node import Node
from reservoirpy.type import NodeInput, Timestep
from reservoirpy.utils.data_validation import check_model_input, check_model_timestep

from .node import Node, OnlineNode, ParallelNode, TrainableNode
from .type import (
    FeedbackBuffers,
    ModelInput,
    ModelTimestep,
    NodeInput,
    State,
    Timeseries,
    Timestep,
    get_data_dimension,
    is_multiseries,
)
from .utils.graphflow import (
    find_indirect_children,
    find_inputs,
    find_outputs,
    find_parents_and_children,
    find_pseudo_inputs,
    topological_sort,
    unique_ordered,
)
from .utils.model_utils import (
    check_input_output_connections,
    check_unnamed_in_out,
    check_unnamed_trainable,
    data_from_buffer,
    join_data,
    mapping_iterator,
    unfold_mapping,
)


class Model:
    """Model base class.

    Parameters
    ----------
    nodes : list of Node, optional
        Nodes to include in the Model.
    edges : list of (Node, Node), optional
        Edges between Nodes in the graph. An edge between a
        Node A and a Node B is created as a tuple (A, B).
    """

    nodes: list[Node]
    edges: list[tuple[Node, int, Node]]
    feedback_buffers: FeedbackBuffers
    inputs: list[Node]
    outputs: list[Node]
    named_nodes: dict[str, Node]
    trainable_nodes: list[Node]
    execution_order: list[Node]
    parents: dict[Node, list[Node]]
    children: dict[Node, list[Node]]

    is_trainable: bool
    is_multi_input: bool
    is_multi_output: bool
    is_online: bool
    is_parallel: bool
    initialized: bool

    def __init__(
        self,
        nodes: Sequence[Node],
        edges: Sequence[tuple[Node, int, Node]],
    ):
        # convert to List[Node], removes duplicates, use dict to preserve order
        self.nodes = unique_ordered(nodes)
        self.edges = unique_ordered(edges)

        self.inputs = find_inputs(self.nodes, self.edges)
        self.outputs = find_outputs(self.nodes, self.edges)
        self.named_nodes = {n.name: n for n in self.nodes if n.name is not None}
        self.trainable_nodes = [n for n in nodes if isinstance(n, TrainableNode)]
        self.is_trainable = len(self.trainable_nodes) > 0
        self.is_multi_input = len(self.inputs) > 1
        self.is_multi_output = len(self.outputs) > 1
        self.is_online = all([isinstance(n, OnlineNode) for n in self.trainable_nodes])
        self.is_parallel = all([isinstance(n, ParallelNode) for n in self.trainable_nodes])
        self.parents, self.children = find_parents_and_children(self.nodes, self.edges)

        # execution order / cycle detection (without teacher forcing)
        direct_edges = [edge for edge in self.edges if edge[1] == 0]
        self.execution_order = topological_sort(self.nodes, direct_edges, inputs=self.inputs)
        self.feedback_buffers = None

        self.initialized = False

    def initialize(
        self,
        x: Union[ModelInput, ModelTimestep],
        y: Optional[Union[ModelInput, ModelTimestep]] = None,
    ):
        """Initializes a :py:class:`Model` instance at runtime, using samples of
        data to infer all :py:class:`Node` dimensions.

        Parameters
        ----------
        x : numpy.ndarray or dict of numpy.ndarray
            A vector of shape `(1, ndim)` corresponding to a timestep of data, or
            a dictionary mapping node names to vector of shapes
            `(1, ndim of corresponding node)`.
        y : numpy.ndarray or dict of numpy.ndarray, optional
            A vector of shape `(1, ndim)` corresponding to a timestep of target
            data, or a dictionary mapping node names to vector of
            shapes `(1, ndim of corresponding node)`.
        """
        check_unnamed_in_out(self)
        check_input_output_connections(self.edges)
        check_unnamed_trainable(self)

        # Turn y into a dict[Node, NodeInput]

        if isinstance(y, Mapping):
            y_: dict[Node, Union[NodeInput, Timestep]] = {self.named_nodes[name]: val for name, val in y.items()}
        elif y is None:
            y_ = {}
        else:
            [trainable_node] = self.trainable_nodes
            y_ = {trainable_node: y}

        # Infer node input dimensions from the input they receive

        node_input_dims = {node: 0 for node in self.nodes}

        if isinstance(x, dict):
            for node_name, val in x.items():
                node = self.named_nodes[node_name]
                node_input_dims[node] += get_data_dimension(val)
        else:
            [node] = self.inputs
            node_input_dims[node] += get_data_dimension(x)

        # also use y as forced teachers. Useful for models with feedback
        indirect_children = find_indirect_children(nodes=self.nodes, edges=self.edges)
        for supervised_node, y_teacher in y_.items():
            for child in indirect_children[supervised_node]:
                node_input_dims[child] += get_data_dimension(y_teacher)

        # execution order / cycle detection (with teacher forcing)
        pseudo_inputs = find_pseudo_inputs(self.nodes, self.edges, y_mapping=y_)
        pseudo_edges = [edge for edge in self.edges if edge[0] not in y_]
        self.pseudo_execution_order = topological_sort(self.nodes, pseudo_edges, inputs=pseudo_inputs)
        # Initialize each node in execution_order
        for node in self.pseudo_execution_order:
            node_input_dim = node_input_dims[node]
            if node.initialized:
                if node_input_dim != node.input_dim:
                    raise ValueError(
                        f"{node} expects input of dimension {node.input_dim} "
                        f"but receives input of dimension {node_input_dim}."
                    )
            else:
                if node in y_:
                    node.initialize(x=np.zeros((node_input_dim,)), y=y_[node])
                else:
                    node.initialize(x=np.zeros((node_input_dim,)))
            if node in y_.keys():
                if get_data_dimension(y_[node]) != node.output_dim:
                    raise ValueError(
                        f"{node} expects training data of dimension {node.output_dim} "
                        f"but receives data of dimension {get_data_dimension(y_[node])}."
                    )
            else:
                for child in self.children[node]:
                    node_input_dims[child] += node.output_dim

        self.feedback_buffers = {(p, d, c): np.zeros((d, p.output_dim)) for p, d, c in self.edges if d > 0}

        # TODO: Jax compilation
        self.initialized = True

    def _step(
        self,
        state: tuple[FeedbackBuffers, Mapping[Node, State]],
        x: Mapping[Node, Timestep],
    ) -> tuple[FeedbackBuffers, dict[Node, State]]:
        buffers, node_states = state

        new_state: dict[Node, State] = {}

        for node in self.execution_order:
            inputs = []
            if node in x:
                inputs.append(x[node])
            inputs += [new_state[parent]["out"] for parent in self.parents[node]]
            inputs += [buffer[-1] for (_p, _d, c), buffer in buffers.items() if c == node]
            node_input = np.concatenate(inputs, axis=-1)
            new_state[node] = node._step(node_states[node], node_input)

        new_buffers = {edge: buffer.copy() for edge, buffer in buffers.items()}
        for (p, d, c), buffer in new_buffers.items():
            buffer[-1] = new_state[p]["out"]
            new_buffers[(p, d, c)] = np.roll(buffer, 1, axis=0)

        return new_buffers, new_state

    def step(self, x: Optional[ModelTimestep] = None) -> ModelTimestep:
        # Auto-regressive mode
        if x is None:
            x = np.zeros((0,))
        check_model_timestep(x)

        if not self.initialized:
            self.initialize(x)

        if isinstance(x, Mapping):
            x_mapping: dict[Node, Timestep] = {self.named_nodes[name]: val for name, val in x.items()}
        else:
            [input_node] = self.inputs
            x_mapping = {input_node: x}

        state = {node: node.state for node in self.nodes}
        buffers = self.feedback_buffers
        new_buffers, new_state = self._step((buffers, state), x_mapping)

        for node in new_state:
            node.state = new_state[node]
        self.feedback_buffers = new_buffers

        if not self.is_multi_output:
            return new_state[self.outputs[0]]["out"]
        else:
            return {node.name: new_state[node]["out"] for node in self.outputs}

    def _run(
        self,
        state: tuple[FeedbackBuffers, Mapping[Node, State]],
        x: Mapping[Node, Timeseries],
    ) -> tuple[tuple[FeedbackBuffers, Mapping[Node, State]], Mapping[Node, Timeseries]]:

        buffers, node_states = state
        # can be run offline (can be faster) if there is no (buffer) feedback
        can_be_run_offline = len(buffers) == 0

        new_state: dict[Node, State] = {}

        if can_be_run_offline:
            output_timeseries: dict[Node, Timeseries] = {}
            # "offline" run: Node by Node
            for node in self.execution_order:
                inputs = []
                if node in x:
                    inputs.append(x[node])
                inputs += [output_timeseries[parent] for parent in self.parents[node]]
                node_input = np.concatenate(inputs, axis=-1)
                new_state[node], output_timeseries[node] = node._run(node_states[node], node_input)
        else:
            # "online" run: step by step
            n_timesteps = x[list(x.keys())[0]].shape[0]
            output_timeseries: dict[Node, Timeseries] = {
                node: np.zeros((n_timesteps, node.output_dim)) for node in self.nodes
            }
            for i, (timestep,) in enumerate(mapping_iterator(x)):
                buffers, node_states = self._step((buffers, node_states), timestep)

                for node in self.nodes:
                    output_timeseries[node][i] = node_states[node]["out"]

        return (buffers, new_state), output_timeseries

    def run(self, x: Optional[ModelInput] = None, iters: Optional[int] = None, workers: int = 1) -> ModelInput:
        # Auto-regressive mode
        if x is None:
            x = np.zeros((iters, 0))
        check_model_input(x)

        if not self.initialized:
            self.initialize(x)

        if isinstance(x, Mapping):
            x_mapping: dict[Node, NodeInput] = {self.named_nodes[name]: val for name, val in x.items()}
        else:
            [input_node] = self.inputs
            x_mapping = {input_node: x}

        previous_states = {node: node.state for node in self.nodes}
        previous_buffers = self.feedback_buffers
        if is_multiseries(x_mapping):
            result: dict[Node, list[Timeseries]] = defaultdict(list)
            iterable_x = unfold_mapping(x_mapping)

            p_output = Parallel(n_jobs=workers, require="sharedmem")(
                delayed(self._run)((previous_buffers, previous_states), timeseries) for timeseries in iterable_x
            )
            new_model_states, output = zip(*p_output)
            new_buffers, new_state = new_model_states[-1]
            for o in output:
                for node in o:
                    result[node].append(o[node])

        else:
            (new_buffers, new_state), result = self._run((previous_buffers, previous_states), x_mapping)

        for node in new_state:
            node.state = new_state[node]
        self.feedback_buffers = new_buffers

        if not self.is_multi_output:
            return result[self.outputs[0]]
        else:
            return {node.name: result[node] for node in self.outputs}

    def predict(self, x: Optional[ModelInput] = None, iters: Optional[int] = None) -> ModelInput:
        return self.run(x=x, iters=iters)

    def _learning_step(
        self,
        state: tuple[FeedbackBuffers, dict[Node, State]],
        x: dict[Node, Timestep],
        y: dict[Node, Timestep],
    ) -> tuple[FeedbackBuffers, dict[Node, State]]:
        buffers, node_states = state
        new_state: dict[Node, State] = {}

        for node in self.execution_order:
            inputs = []
            if node in x:
                inputs.append(x[node])
            inputs += [buffer[-1] for (_p, _d, c), buffer in buffers.items() if c == node]
            inputs += [new_state[parent]["out"] for parent in self.parents[node]]
            node_input = np.concatenate(inputs, axis=-1)
            if isinstance(node, OnlineNode):
                node_target = y.get(node, None)
                new_state[node] = {"out": node._learning_step(node_input, node_target)}
            else:
                new_state[node] = node._step(node_states[node], node_input)

        new_buffers = {edge: buffer.copy() for edge, buffer in buffers.items()}
        for (p, d, c), buffer in new_buffers.items():
            buffer[-1] = new_state[p]["out"]
            new_buffers[(p, d, c)] = np.roll(buffer, 1, axis=0)

        return new_buffers, new_state

    def partial_fit(
        self,
        x: Union[Timeseries, dict[str, Timeseries]],
        y: Optional[Union[Timeseries, dict[str, Timeseries]]] = None,
    ) -> ModelInput:
        if not self.is_online:
            raise TypeError("Trying to partial_fit a Model that can't be trained in an online manner.")

        check_model_input(x)
        if y is not None:
            check_model_input(y)

        if not self.initialized:
            self.initialize(x, y)

        # Turn x into a dict[Node, NodeInput]
        if isinstance(x, Mapping):
            x_ = {self.named_nodes[node]: value for node, value in x.items()}
        else:
            [input_node] = self.inputs
            x_ = {input_node: x}

        # Turn y into a dict[Node, NodeInput]
        if y is None:
            y_ = {}
        elif isinstance(y, Mapping):
            y_ = {self.named_nodes[name]: val for name, val in y.items()}
        else:
            [trainable_node] = self.trainable_nodes
            y_ = {trainable_node: y}

        n_timesteps = x_[list(x_.keys())[0]].shape[0]
        output_timeseries: dict[Node, Timeseries] = {
            node: np.zeros((n_timesteps, node.output_dim)) for node in self.nodes
        }

        states = {node: node.state for node in self.nodes}
        buffers = self.feedback_buffers
        for i, (x_timestep, y_timestep) in enumerate(mapping_iterator(x_, y_)):
            buffers, states = self._learning_step((buffers, states), x_timestep, y_timestep)
            for node in self.nodes:
                output_timeseries[node][i] = states[node]["out"]

        for node in states:
            node.state = states[node]
        self.feedback_buffers = buffers

        if not self.is_multi_output:
            return output_timeseries[self.outputs[0]]
        else:
            return {node.name: output_timeseries[node] for node in self.outputs}

    def fit(
        self,
        x: ModelInput,
        y: Optional[ModelInput] = None,
        warmup: int = 0,
        workers: int = 1,
    ) -> "Model":
        check_model_input(x)
        if y is not None:
            check_model_input(y)

        if not self.initialized:
            self.initialize(x, y)

        # TODO: try removing the default list
        result: dict[Node, NodeInput] = defaultdict(list)
        buffers = self.feedback_buffers

        # Turn x into a dict[Node, NodeInput]
        if isinstance(x, Mapping):
            x_ = {self.named_nodes[node]: value for node, value in x.items()}
        else:
            [input_node] = self.inputs
            x_ = {input_node: x}

        # Turn y into a dict[Node, NodeInput]
        if y is None:
            y_ = {}
        elif isinstance(y, Mapping):
            y_ = {self.named_nodes[name]: val for name, val in y.items()}
        else:
            [trainable_node] = self.trainable_nodes
            y_ = {trainable_node: y}

        # forced teaching
        for supervised in y_:
            # TODO: handle Unsupervised has children
            result[supervised] = y_[supervised]

        for node in self.pseudo_execution_order:
            inputs: list[NodeInput] = []
            if node in x_:
                inputs.append(x_[node])
            inputs += [result[parent] for parent in self.parents[node]]
            for (p, _d, c), buffer in buffers.items():
                if c == node:
                    new_buffer, data = data_from_buffer(buffer, result[p])
                    buffers[(p, _d, c)] = new_buffer
                    inputs.append(data)
            node_input = join_data(*inputs)
            if isinstance(node, TrainableNode):
                node_target = y_.get(node, None)
                if isinstance(node, ParallelNode):
                    node.fit(node_input, node_target, warmup=warmup, workers=workers)
                else:
                    node.fit(node_input, node_target, warmup=warmup)
            else:
                result[node] = node.run(node_input, workers=workers)

        return self

    def __call__(self, x: Optional[ModelTimestep] = None) -> ModelTimestep:
        return self.step(x)

    def __rshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link

        return link(self, other)

    def __rrshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link

        return link(other, self)

    def __lshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=other, receiver=self)

    def __rlshift__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import link_feedback

        return link_feedback(sender=self, receiver=other)

    def __and__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import merge

        return merge(self, other)

    def __rand__(self, other: Union[Node, "Model", Sequence[Union[Node, "Model"]]]) -> "Model":
        from .ops import merge

        return merge(other, self)



# Author: Deepayan Das at 16/08/2023 <deepayan.das@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from copy import deepcopy
from typing import Any, Optional, Sequence, Union

import numpy as np

from reservoirpy.type import NodeInput, State, Timeseries, Timestep, is_array
from reservoirpy.utils.data_validation import check_node_input

from ..node import TrainableNode
from ..utils.random import rand_generator


class ScikitLearnNode(TrainableNode):
    """
    A node interfacing a scikit-learn linear model that can be used as an offline
    readout node.

    The ScikitLearnNode takes a scikit-learn model as parameter and creates a
    node with the specified model.

    We currently support classifiers (like
    :py:class:`sklearn.linear_model.LogisticRegression` or
    :py:class:`sklearn.linear_model.RidgeClassifier`) and regressors (like
    :py:class:`sklearn.linear_model.Lasso` or
    :py:class:`sklearn.linear_model.ElasticNet`).

    For more information on the above-mentioned estimators,
    please visit scikit-learn linear model API reference
    <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.linear_model>`_


    Parameters
    ----------
    model : class, scikit-learn model
        scikit-learn class to be wrapped by the Node.
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    name : str, optional
        Node name.
    **kwargs
        Additional keyword arguments passed to the scikit-learn model.

    Example
    -------
    >>> from reservoirpy.nodes import Reservoir, ScikitLearnNode
    >>> from sklearn.linear_model import Lasso
    >>> reservoir = Reservoir(units=100)
    >>> readout = ScikitLearnNode(model=Lasso, model_hypers={"alpha":1e-5})
    >>> model = reservoir >> readout
    """

    #: scikit-learn class to be wrapped by the Node.
    model: type["sklearn.base.BaseEstimator"]
    #: Additional keyword arguments passed to the scikit-learn model.
    model_kwargs: dict[str, Any]
    #: Model instance or list of model instances if multiple output are expected
    #: and the scikit-learn model doesn't support it.
    instances: Union["sklearn.base.BaseEstimator", list["sklearn.base.BaseEstimator"]]

    def __init__(
        self,
        model,
        output_dim: Optional[int] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        if not hasattr(model, "fit"):
            model_name = model.__name__
            raise AttributeError(f"Specified model {model_name} has no method called 'fit'.")
        if not hasattr(model, "predict"):
            model_name = model.__name__
            raise AttributeError(f"Specified model {model_name} has no method called 'predict'.")

        # Ensure reproducibility
        # scikit-learn currently only supports RandomState
        if not "random_state" in kwargs and "random_state" in model.__init__.__kwdefaults__:
            generator = rand_generator()
            kwargs.update({"random_state": np.random.RandomState(seed=generator.integers(1 << 32))})

        self.model = model
        self.name = name
        self.model_kwargs = kwargs
        self.output_dim = output_dim
        self.state = {"out": None}
        self.initialized = False

    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        self._set_input_dim(x)
        self._set_output_dim(y)

        # TODO: Use https://scikit-learn.org/stable/modules/generated/sklearn.multioutput.MultiOutputRegressor.html instead
        first_instance = self.model(**deepcopy(self.model_kwargs))
        # If there are multiple output but the specified model doesn't support
        # multiple outputs, we create an instance of the model for each output.
        if self.output_dim > 1 and not first_instance._get_tags().get("multioutput"):
            self.instances = [self.model(**deepcopy(self.model_kwargs)) for i in range(self.output_dim)]
        else:
            self.instances = first_instance
        self.state = {"out": np.zeros((self.output_dim,))}
        self.initialized = True

    def _step(self, state: State, x: Timestep):
        instances = self.instances
        if not isinstance(instances, list):
            res = instances.predict(x.reshape(1, -1)).ravel()
        else:
            res = np.concatenate(
                [instance.predict(x.reshape(1, -1)).ravel() for instance in instances],
                axis=-1,
            )
        return {"out": res}

    def _run(self, state: State, x: Timeseries) -> tuple[State, Timeseries]:
        instances = self.instances
        if not isinstance(instances, list):
            res = instances.predict(x)
            if res.ndim == 1:
                res = res[..., np.newaxis]
        else:
            res = np.column_stack([instance.predict(x) for instance in instances])
        return {"out": res[-1]}, res

    def fit(self, x: NodeInput, y: Optional[NodeInput] = None, warmup: int = 0):
        check_node_input(x, expected_dim=self.input_dim)
        if y is not None:
            check_node_input(y, expected_dim=self.output_dim)

        if not self.initialized:
            self.initialize(x, y)

        if isinstance(x, Sequence):
            # Concatenate all the batches as one np.ndarray
            # of shape (timeseries*timesteps, features)
            x = np.concatenate(x, axis=0)
        if is_array(x) and x.ndim == 3:
            x = x.reshape(-1, x.shape[-1])

        if isinstance(y, Sequence):
            # Concatenate all the batches as one np.ndarray
            # of shape (timeseries*timesteps, features)
            y = np.concatenate(y, axis=0)
        if is_array(y) and y.ndim == 3:
            y = y.reshape(-1, y.shape[-1])

        instances = self.instances
        if not isinstance(instances, list):
            if self.output_dim > 1:
                # Multi-output node and multi-output sklearn model
                instances.fit(x, y)
            else:
                # Y_ should have 1 feature so we reshape to
                # (timeseries, ) to avoid scikit-learn's DataConversionWarning
                instances.fit(x, y[..., [0]])
        else:
            for i, instance in enumerate(instances):
                instance.fit(x, y[..., [i]])

        return self



# Author: Nathan Trouvain at 15/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from reservoirpy.nodes import Identity, ReLU, Sigmoid, Softmax, Softplus, Tanh


@pytest.mark.parametrize(
    "node",
    (Tanh(), Softmax(), Softplus(), Sigmoid(), Identity(), ReLU(), Softmax(beta=2.0)),
)
def test_activation_step(node):
    x = np.ones((10,))
    out = node(x)
    assert out.shape == x.shape
    assert node.input_dim == x.shape[-1]
    assert node.output_dim == x.shape[-1]


@pytest.mark.parametrize(
    "node",
    (Tanh(), Softmax(), Softplus(), Sigmoid(), Identity(), ReLU(), Softmax(beta=2.0)),
)
def test_activation_run(node):
    x = np.ones((100, 10))
    out = node.run(x)
    assert out.shape == x.shape
    assert node.input_dim == x.shape[-1]
    assert node.output_dim == x.shape[-1]

    out2 = node.step(x[-1])
    np.testing.assert_array_almost_equal(out[-1], out2)

    X = list(np.ones((7, 100, 10)))
    out = node.run(X)
    assert isinstance(out, list)
    assert out[0].shape == X[0].shape
    assert node.input_dim == X[0].shape[-1]
    assert node.output_dim == X[0].shape[-1]

    out2 = node.step(X[0][-1])
    np.testing.assert_array_almost_equal(out[0][-1], out2)



# Author: Nathan Trouvain at 06/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial
from typing import Callable, Optional, Sequence, Tuple, Union

from ..activationsfunc import identity, relu, sigmoid, softmax, softplus, tanh
from ..node import Node
from ..type import NodeInput, State, Timeseries, Timestep


class F(Node):
    """Generic elementwise and stateless Node

    This Node takes a function f: array -> array that can take both timeseries
    and timesteps.

    Parameters
    ----------
    f: callable
        Beta parameter of softmax.
    name : str, optional
        Node name.
    **kwargs :
        Additional arguments passed to f.
    """

    def __init__(self, f: Callable, name: Optional[str] = None, **kwargs):
        self.f = partial(f, **kwargs)
        self.state = {}
        self.initialized = False
        self.name = name

    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: None = None,
    ):
        self._set_input_dim(x)
        self.output_dim = self.input_dim
        self.initialized = True

    def _step(self, state: tuple, x: Timestep) -> State:
        return {"out": self.f(x)}

    def _run(self, state: State, x: Timeseries) -> Tuple[State, Timeseries]:
        out = self.f(x)
        return {"out": out[-1]}, out


class Softmax(F):
    """Softmax activation function.

    .. math::

        y_k = \\frac{e^{\\beta x_k}}{\\sum_{i=1}^{n} e^{\\beta x_i}}

    Parameters
    ----------
    beta: float, default to 1.0
        Beta parameter of softmax.
    name : str, optional
        Node name.
    """

    def __init__(self, beta=1.0, name: Optional[str] = None):
        self.f = partial(softmax, beta=beta)
        self.state = {}
        self.initialized = False
        self.name = name


class Softplus(F):
    """Softplus activation function.

    .. math::

        f(x) = \\mathrm{ln}(1 + e^{x})

    Parameters
    ----------
    name : str, optional
        Node name.
    """

    def __init__(self, name: Optional[str] = None):
        self.f = softplus
        self.state = {}
        self.initialized = False
        self.name = name


class Sigmoid(F):
    """Sigmoid activation function.

    .. math::

        f(x) = \\frac{1}{1 + e^{-x}}

    Parameters
    ----------
    name : str, optional
        Node name.
    """

    def __init__(self, name: Optional[str] = None):
        self.f = sigmoid
        self.state = {}
        self.initialized = False
        self.name = name


class Tanh(F):
    """Hyperbolic tangent activation function.

    .. math::

        f(x) = \\frac{e^x - e^{-x}}{e^x + e^{-x}}

    Parameters
    ----------
    name : str, optional
        Node name.
    """

    def __init__(self, name: Optional[str] = None):
        self.f = tanh
        self.state = {}
        self.initialized = False
        self.name = name


class Identity(F):
    """Identity function.

    .. math::

        f(x) = x

    Provided for convenience.

    Parameters
    ----------
    name : str, optional
        Node name.
    """

    def __init__(self, name: Optional[str] = None):
        self.f = identity
        self.state = {}
        self.initialized = False
        self.name = name


class ReLU(F):
    """ReLU activation function.

    .. math::

        f(x) = x ~~ \\mathrm{if} ~~ x > 0 ~~ \\mathrm{else} ~~ 0

    Parameters
    ----------
    name : str, optional
        Node name.
    """

    def __init__(self, name: Optional[str] = None):
        self.f = relu
        self.state = {}
        self.initialized = False
        self.name = name



from typing import Callable, Optional, Sequence, Union

import numpy as np

from reservoirpy.utils.data_validation import check_node_input

from ..mat_gen import zeros
from ..node import OnlineNode
from ..type import NodeInput, State, Timeseries, Timestep, Weights, is_array


class LMS(OnlineNode):
    """Single layer of neurons learning connections using Least Mean Squares
    algorithm.

    The learning rules is well described in [1]_.

    Parameters
    ----------
    learning_rate : float or Python generator or iterable, default to 1e-6
        Learning rate. If an iterable or a generator is provided, the learning rate can
        be changed at each timestep of training. A new learning rate will be drawn from
        the iterable or generator at each timestep.
    Wout : callable or array-like of shape (units, targets), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    fit_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    input_dim : int, optional
        Number of input dimensions in the readout, can be inferred at first call.
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    name : str, optional
        Node name.

    Examples
    --------
    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.01, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.

    >>> from reservoirpy.nodes import LMS
    >>> lms_node = LMS(alpha=1e-1)

    >>> lms_node.train(x[:50], y[:50])
    >>> print(lms_node.Wout.T, lms_node.bias)
    [[ 9.156 -0.967   6.411]] [[11.564]]
    >>> lms_node.train(x[50:], y[50:])
    >>> print(lms_node.Wout.T, lms_node.bias)
    [[ 9.998 -0.202  7.001]] [[12.005]]

    References
    ----------

    .. [1] Sussillo, D., & Abbott, L. F. (2009). Generating Coherent Patterns of
           Activity from Chaotic Neural Networks. Neuron, 63(4), 544–557.
           https://doi.org/10.1016/j.neuron.2009.07.018
    """

    #: Learned output weights (:math:`\mathbf{W}_{out}`).
    Wout: Weights
    #: Learned bias (:math:`\mathbf{b}`).
    bias: Weights
    #: Learning rate (:math:`\alpha`) (:math:`1\cdot 10^{-6}` by default).
    learning_rate: float
    #: If True, learn a bias term (True by default).
    fit_bias: bool

    def __init__(
        self,
        learning_rate: float = 1e-6,
        Wout: Union[Weights, Callable] = zeros,
        bias: Union[Weights, Callable] = zeros,
        fit_bias: bool = True,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.learning_rate = learning_rate
        self.Wout = Wout
        self.bias = bias
        self.fit_bias = fit_bias
        self.initialized = False
        self.state = {}
        self.name = name

        # set input_dim/output_dim (if possible)
        self.input_dim = input_dim
        self.output_dim = output_dim
        if is_array(Wout):
            if input_dim is not None and Wout.shape[0] != input_dim:
                raise ValueError(
                    f"Both 'input_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {input_dim} != {Wout.shape[0]}."
                )
            self.input_dim = Wout.shape[0]
            if output_dim is not None and Wout.shape[1] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {Wout.shape[1]}."
                )
            self.output_dim = Wout.shape[1]
        if is_array(bias):
            if output_dim is not None and bias.shape[0] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {bias.shape[0]}."
                )
            self.output_dim = bias.shape[0]

    def _run(self, state: tuple, x: Timeseries) -> tuple[State, Timeseries]:
        out = x @ self.Wout + self.bias
        return {"out": out[-1]}, out  # (len, in) @ (in, out) + (out,)

    def _step(self, state: tuple, x: Timestep) -> State:
        return {"out": x @ self.Wout + self.bias}  # (in, ) @ (in, out) + (out,)

    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        self._set_input_dim(x)
        self._set_output_dim(y)

        # initialize matrices
        if isinstance(self.Wout, Callable):
            self.Wout = self.Wout(self.input_dim, self.output_dim)
        if isinstance(self.bias, Callable):
            self.bias = self.bias(self.output_dim)

        self.initialized = True

    def _learning_step(self, x: Timestep, y: Timestep) -> Timestep:
        alpha: float = self.learning_rate
        Wout: Weights = self.Wout
        bias: Weights = self.bias

        prediction = x @ Wout + bias  # (out,) = (in,) @ (in, out) + (out,)
        error = prediction - y  # (out,)
        dWout = -alpha * np.outer(x, error)  # (in, out)
        Wout_next = Wout + dWout  # (in, out)
        if self.fit_bias:
            dbias = -alpha * error
            bias_next = bias + dbias
        else:
            bias_next = bias
        y_pred = x @ Wout_next + bias

        self.Wout = Wout_next
        self.bias = bias_next
        return y_pred

    def partial_fit(self, x: Timeseries, y: Timeseries):
        check_node_input(x, expected_dim=self.input_dim)
        if y is not None:
            check_node_input(y)

        if not self.initialized:
            self.initialize(x, y)

        n_timesteps = x.shape[-2]
        out_dim = y.shape[-1]
        y_pred = np.empty((n_timesteps, out_dim))
        for i, (x_, y_) in enumerate(zip(x, y)):
            y_pred_ = self._learning_step(x_, y_)
            y_pred[i] = y_pred_

        self.state = {"out": y_pred_}
        return y_pred



# Author: Nathan Trouvain at 06/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Optional, Union

import numpy as np
from numpy.random import Generator, default_rng

__SEED = None
__global_rg = default_rng()


def set_seed(seed):
    """Set random state seed globally.

    Parameters
    ----------
        seed : int
    """
    global __SEED
    global __global_rg
    if type(seed) is int:
        __SEED = seed
        __global_rg = default_rng(__SEED)
        np.random.seed(__SEED)
    else:
        raise TypeError(f"Random seed must be an integer, not {type(seed)}")


def rand_generator(seed: Optional[Union[int, Generator]] = None) -> Generator:
    if seed is None:
        return __global_rg.spawn(n_children=1)[0]
    if isinstance(seed, Generator):
        return seed
    else:
        return default_rng(seed)


def noise(rng, dist="normal", shape=1, gain=1.0, **kwargs):
    """Generate noise from a given distribution, and apply a gain factor.

    Parameters
    ----------
        rng : numpy.random.Generator
            A random number generator.
        dist : str, default to 'normal'
            A random variable distribution.
        shape : int or tuple of ints, default to 1
            Shape of the noise vector.
        gain : float, default to 1.0
            Gain factor applied to noise.
        **kwargs
            Any other parameters of the noise distribution.

    Returns
    -------
        np.ndarray
            A noise vector.

    Note
    ----
        If `gain` is 0, then noise vector is null.
    """
    if abs(gain) > 0.0:
        return gain * getattr(rng, dist)(**kwargs, size=shape)
    else:
        return np.zeros(shape)



"""
=========================================================
Node operations - link, merge (:py:mod:`reservoirpy.ops`)
=========================================================

Operations on :py:class:`~.Node` and :py:class:`~.Model`.

.. currentmodule:: reservoirpy.ops

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   link - Link Nodes or Model with a direct connection.
   merge - Merge Models.
   link_feedback - Link Nodes or Models with a feedback connection.
"""

from itertools import product
from typing import Sequence, Union

# Author: Nathan Trouvain at 25/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from reservoirpy.node import Node

from .model import Model
from .node import Node
from .utils.graphflow import unique_ordered


def _check_all_models(*operands):
    msg = "Impossible to link nodes: object {} is neither a Node nor a Model."
    for operand in operands:
        if isinstance(operand, Sequence):
            for model in operand:
                if not (isinstance(model, Node) or isinstance(model, Model)):
                    raise TypeError(msg.format(model))
        elif not (isinstance(operand, Node) or isinstance(operand, Model)):
            raise TypeError(msg.format(operand))


def _link_1to1(
    left: Union[Node, Model], right: Union[Node, Model], delay=0
) -> tuple[list[Node], list[tuple[Node, int, Node]]]:
    """Connects two nodes or models. See `link` doc for more info."""
    # fetch all nodes in the two subgraphs, if they are models.

    nodes_left = [left] if isinstance(left, Node) else left.nodes
    nodes_right = [right] if isinstance(right, Node) else right.nodes
    all_nodes = unique_ordered(nodes_left + nodes_right)

    # fetch all edges in the two subgraphs, if they are models.
    edges_left = left.edges if isinstance(left, Model) else []
    edges_right = right.edges if isinstance(right, Model) else []
    all_edges = unique_ordered(edges_left + edges_right)

    # create edges between output nodes of the
    # subgraph 1 and input nodes of the subgraph 2.
    senders = [left] if isinstance(left, Node) else left.outputs
    receivers = [right] if isinstance(right, Node) else right.inputs

    new_edges = list(product(senders, [delay], receivers))

    # maybe nodes are already initialized?
    # check if connected dimensions are ok
    for sender, _, receiver in new_edges:
        if sender.initialized and receiver.initialized and sender.output_dim != receiver.input_dim:
            raise ValueError(
                f"Dimension mismatch between connected nodes: "
                f"sender node {sender} has output dimension "
                f"{sender.output_dim} but receiver node "
                f"{receiver} has input dimension "
                f"{receiver.input_dim}."
            )

    # all outputs from subgraph 1 are connected to
    # all inputs from subgraph 2.
    return all_nodes, all_edges + new_edges


def link(
    left: Union[Node, Model, Sequence[Union[Node, Model]]],
    right: Union[Node, Model, Sequence[Union[Node, Model]]],
) -> Model:
    """Link two :py:class:`~.Node` instances to form a :py:class:`~.Model`
    instance. `node1` output will be used as input for `node2` in the
    created model. This is similar to a function composition operation:

    .. math::

        model(x) = (node1 \\circ node2)(x) = node2(node1(x))

    You can also perform this operation using the ``>>`` operator::

        model = node1 >> node2

    Or using this function::

        model = link(node1, node2)

    -`node1` and `node2` can also be :py:class:`~.Model` instances. In this
    case, the new :py:class:`~.Model` created will contain all nodes previously
    contained in all the models, and link all `node1` outputs to all `node2`
    inputs. This allows to chain the  ``>>`` operator::

        step1 = node0 >> node1  # this is a model
        step2 = step1 >> node2  # this is another

    -`node1` and `node2` can finally be lists or tuples of nodes. In this
    case, all `node1` outputs will be linked to all `node2` inputs. You can
    still use the ``>>`` operator in this situation, except for many-to-many
    nodes connections::

        # many-to-one
        model = [node1, node2, ..., node] >> node_out
        # one-to-many
        model = node_in >> [node1, node2, ..., node]
        # ! many-to-many requires to use the `link` method explicitly!
        model = link([node1, node2, ..., node], [node1, node2, ..., node])

    Parameters
    ----------
        left : Node, Model or list of Node
            Nodes or lists of nodes to link.
        right : Node, Model or list of Node
            Nodes or lists of nodes to link.

    Returns
    -------
        Model
            A :py:class:`~.Model` instance chaining the nodes.

    Raises
    ------
        TypeError
            Dimension mismatch between connected nodes: `left` output
            dimension if different from `right` input dimension.
            Reinitialize the nodes or create new ones.

    Notes
    -----

        Be careful to how you link the different nodes: `reservoirpy` does not
        allow to have circular dependencies between them::

            model = node1 >> node2  # fine
            model = node1 >> node2 >> node1  # raises! data would flow in
                                             # circles forever...
    """
    # TODO: copy delay buffers
    _check_all_models(left, right)

    left_seq: Sequence = left if isinstance(left, Sequence) else [left]
    right_seq: Sequence = right if isinstance(right, Sequence) else [right]

    nodes: list[Node] = []
    edges: list[tuple[Node, int, Node]] = []

    for left_node in left_seq:
        for right_node in right_seq:
            new_nodes, new_edges = _link_1to1(left_node, right_node)
            nodes += new_nodes
            edges += new_edges

    return Model(nodes=unique_ordered(nodes), edges=unique_ordered(edges))


def merge(*models: Union[Node, Model, Sequence[Union[Node, Model]]]) -> Model:
    """Merge different :py:class:`~.Model` or :py:class:`~.Node`
    instances into a single :py:class:`~.Model` instance.

    :py:class:`~.Node` instances contained in the models to merge will be
    gathered in a single model, along with all previously defined connections
    between them, if they exists.

    You can also perform this operation using the ``&`` operator::

        model = (node1 >> node2) & (node1 >> node3))

    This is equivalent to::

        model = merge((node1 >> node2), (node1 >> node3))

    The in-place operator can also be used::

        model &= other_model

    Which is equivalent to::

        model.update_graph(other_model.nodes, other_model.edges)

    Parameters
    ----------
    model: Model or Node
        First node or model to merge.
    *models : Model or Node
        All models to merge.

    Returns
    -------
    Model
        A new :py:class:`~.Model` instance.

    """
    # TODO: copy delay buffers
    _check_all_models(*models)

    nodes: list[Node] = []
    edges: list[tuple[Node, int, Node]] = []

    for model in models:
        if isinstance(model, Sequence):
            for element in model:
                if isinstance(element, Node):
                    nodes.append(element)
                elif isinstance(element, Model):
                    nodes += element.nodes
                    edges += element.edges
                else:
                    TypeError(f"Impossible to merge models: object {type(model)} is not a Node or a Model.")
        elif isinstance(model, Node):
            nodes.append(model)
        elif isinstance(model, Model):
            nodes += model.nodes
            edges += model.edges
        else:
            TypeError(f"Impossible to merge models: object {type(model)} is not a Node or a Model.")

    return Model(nodes=unique_ordered(nodes), edges=unique_ordered(edges))


def link_feedback(
    sender: Union[Node, Model, Sequence[Union[Node, Model]]],
    receiver: Union[Node, Model, Sequence[Union[Node, Model]]],
) -> Model:
    """Create a feedback connection from the ``sender`` to the ``receiver``.
    Feedback connections are regular node-to-node connections with a delay of
    one timestep.

    If ``sender`` or ``receiver`` is a Model or a list of Nodes or Models, all
    outputs of ``sender`` will be connected as feedback to ``receiver``.

     You can also perform this operation using the ``<<`` operator::

        node1 = node1 << node2
        # with feedback from a Model
        node1 = node1 << (fbnode1 >> fbnode2)
        # with feedback from a list of nodes or models
        node1 = node1 << [fbnode1, fbnode2, ...]

    You can also use this function to define feedback::

        node1 = link_feedback(node1, node2)

    Parameters
    ----------
    sender : Node, Model, or list of Nodes or Models
        Node(s) or Model(s) sending feedback.
    receiver : Node, Model, or list of Nodes or Models
        Node(s) or Model(s) receiving feedback.

    Returns
    -------
        Model
            A model instance with all node and connections from ``sender`` and
            ``receiver`` and with feedback connections from ``sender`` to
            ``receiver``.

    Raises
    ------
        TypeError
            If any of the senders or receivers are not Nodes or Models.
    """
    # TODO: copy delay buffers
    _check_all_models(sender, receiver)

    left_seq: Sequence = sender if isinstance(sender, Sequence) else [sender]
    right_seq: Sequence = receiver if isinstance(receiver, Sequence) else [receiver]

    nodes: list[Node] = []
    edges: list[tuple[Node, int, Node]] = []

    for sender_element in left_seq:
        for receiver_element in right_seq:
            new_nodes, new_edges = _link_1to1(sender_element, receiver_element, delay=1)
            nodes += new_nodes
            edges += new_edges

    return Model(nodes=unique_ordered(nodes), edges=unique_ordered(edges))



# Author: Nathan Trouvain at 05/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from reservoirpy import ESN
from reservoirpy.mat_gen import normal
from reservoirpy.nodes import Reservoir, Ridge


def test_esn_init():
    esn = ESN(units=100, lr=0.8, sr=0.4, ridge=1e-5)

    data = np.ones((1000, 10))
    res = esn.fit(data, data)
    data = np.ones((10,))
    res = esn(data)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 10)
    assert esn.reservoir.lr == 0.8
    assert esn.reservoir.units == 100

    data = np.ones((1000, 10))
    res = esn.run(data)
    assert res.shape == (1000, 10)

    esn = ESN(units=100, lr=0.8, sr=0.4, ridge=1e-5, output_dim=7)
    with pytest.raises(ValueError):
        esn.initialize(np.ones((10, 3)), np.ones((10, 8)))


def test_esn_init_from_obj():
    res = Reservoir(100, lr=0.8, sr=0.4)
    read = Ridge(ridge=1e-5, output_dim=1)

    esn = ESN(reservoir=res, readout=read)

    x = np.ones((10, 2))
    y = np.arange(10).reshape(-1, 1)
    esn.fit(x, y)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 2)
    assert esn.reservoir.lr == 0.8
    assert esn.reservoir.units == 100

    data = np.ones((1000, 2))
    res = esn.run(data)

    assert res.shape == (1000, 1)


# def test_esn_states(): # TODO: implement a way to return reservoir activity
#     res = Reservoir(100, lr=0.8, sr=0.4)
#     read = Ridge(ridge=1e-5, output_dim=1)

#     esn = ESN(reservoir=res, readout=read)

#     data = np.ones((2, 10, 10))
#     out = esn.run(data, return_states="all")

#     assert out["reservoir"][0].shape == (10, 100)
#     assert out["readout"][0].shape == (10, 1)

#     out = esn.run(data, return_states=["reservoir"])

#     assert out["reservoir"][0].shape == (10, 100)

#     s_reservoir = esn.state()
#     assert_equal(s_reservoir, res.state())

#     s_readout = esn.state(which="readout")
#     assert_equal(s_readout, read.state())

#     with pytest.raises(ValueError):
#         esn.state(which="foo")


def test_esn_feedback():
    esn = ESN(units=100, output_dim=5, lr=0.8, sr=0.4, ridge=1e-5, feedback=True)

    x = np.ones((13, 10))
    y = np.ones((13, 5))
    res = esn.fit(x, y).run(x)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 15)
    assert esn.readout.Wout.shape == (100, 5)
    assert res.shape == (13, 5)

    res = esn(x[0])


def test_esn_argument_collision():
    esn = ESN(
        units=100,
        input_dim=15,  # should be the input_dim of the reservoir, not the readout
        output_dim=5,
        lr=0.8,
        sr=0.4,
        bias=1.0,  # should be the bias of the reservoir, not the readout
        readout_bias=normal,  # should be the bias of the readout
        fit_bias=False,
        seed=1,
        name="MyNode",
        ridge=1e-5,
        feedback=True,
        input_to_readout=True,
    )

    x = np.ones((13, 10))
    y = np.ones((13, 5))
    res = esn.fit(x, y).run(x)

    assert esn.reservoir.output_dim == 100
    assert esn.reservoir.input_dim == 15
    assert esn.readout.input_dim == 110
    assert esn.readout.output_dim == 5
    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 15)
    assert esn.readout.Wout.shape == (110, 5)
    assert esn.readout.bias.shape == (5,)
    assert esn.reservoir.bias == 1.0
    assert res.shape == (13, 5)
    assert len(esn.nodes) == 3
    assert len(esn.edges) == 4

    res = esn(x[0])


def test_esn_freerunning():
    rng = np.random.default_rng(seed=1)
    model = ESN(units=100, sr=1.0, bias=normal)
    x, y = rng.normal(size=(100, 0)), rng.normal(size=(100, 3))
    model.fit(x, y)
    res = model.predict(x=None, iters=213)

    assert model.reservoir.input_dim == 0
    assert model.readout.input_dim == 100
    assert model.readout.output_dim == 3
    assert res.shape == (213, 3)



import numpy as np
import pytest
from numpy.random import default_rng
from numpy.testing import (
    assert_allclose,
    assert_array_almost_equal,
    assert_array_equal,
    assert_raises,
)
from scipy import linalg, sparse

from reservoirpy.mat_gen import (
    bernoulli,
    cluster,
    fast_spectral_initialization,
    line,
    normal,
    ones,
    orthogonal,
    random_sparse,
    ring,
    small_world,
    uniform,
    zeros,
)


@pytest.mark.parametrize(
    "shape,dist,connectivity,kwargs,expects",
    [
        ((50, 50), "uniform", 0.1, {}, "sparse"),
        ((50, 50), "uniform", 0.1, {"loc": 5.0, "scale": 1.0}, "sparse"),
        ((50, 50), "uniform", 1.0, {}, "dense"),
        ((50, 50), "custom_bernoulli", 0.1, {}, "sparse"),
        ((50, 50, 50), "custom_bernoulli", 0.1, {"p": 0.9}, "dense"),
        ((50, 50), "custom_bernoulli", 1.0, {}, "dense"),
        ((50, 50), "foo", 0.1, {}, "raise"),
        ((50, 50), "uniform", 5.0, {}, "raise"),
        ((50, 50), "uniform", 0.1, {"p": 0.9}, "raise"),
        ((50, 5), "uniform", 0.1, {"degree": 23, "direction": "out"}, "sparse"),
        ((50, 5), "uniform", 0.1, {"degree": 3, "direction": "in"}, "sparse"),
        ((50, 5), "uniform", 0.1, {"degree": 6, "direction": "in"}, "raise"),
        ((50, 5), "uniform", 0.1, {"degree": -1000, "direction": "out"}, "raise"),
    ],
)
def test_random_sparse(shape, dist, connectivity, kwargs, expects):
    if expects == "sparse":
        init = random_sparse(dist=dist, connectivity=connectivity, seed=42, **kwargs)
        w0 = init(*shape)
        w1 = random_sparse(
            *shape, dist=dist, connectivity=connectivity, seed=42, **kwargs
        )

        w0 = w0.toarray()
        w1 = w1.toarray()

    if expects == "dense":
        init = random_sparse(dist=dist, connectivity=connectivity, seed=42, **kwargs)
        w0 = init(*shape)
        w1 = random_sparse(
            *shape, dist=dist, connectivity=connectivity, seed=42, **kwargs
        )

    if expects == "raise":
        with pytest.raises(Exception):
            init = random_sparse(
                dist=dist, connectivity=connectivity, seed=42, **kwargs
            )
            w0 = init(*shape)
        with pytest.raises(Exception):
            w1 = random_sparse(
                *shape, dist=dist, connectivity=connectivity, seed=42, **kwargs
            )
    else:
        assert_array_equal(w1, w0)
        if kwargs.get("degree") is None:
            assert_allclose(np.count_nonzero(w0) / w0.size, connectivity, atol=1e-2)
        else:
            dim_length = {"in": shape[0], "out": shape[1]}
            assert (
                np.count_nonzero(w0)
                == kwargs["degree"] * dim_length[kwargs["direction"]]
            )


@pytest.mark.parametrize(
    "shape,sr,input_scaling,kwargs,expects",
    [
        ((50, 50), 2.0, None, {"connectivity": 0.1}, "sparse"),
        ((50, 50), None, -2.0, {"connectivity": 1.0}, "dense"),
        ((50, 50), 2.0, None, {"connectivity": 1.0}, "dense"),
        ((50, 50), None, -2.0, {"connectivity": 1.0}, "dense"),
        ((50, 50), None, np.ones((50,)) * 0.1, {"connectivity": 1.0}, "dense"),
        ((50, 50), None, np.ones((50,)) * 0.1, {"connectivity": 0.1}, "sparse"),
        ((50, 50), 2.0, -2.0, {"connectivity": 0.1}, "raise"),
        ((50, 50), None, 1e-12, {"connectivity": 0.1}, "sparse"),
    ],
)
def test_random_sparse_scalings(shape, sr, input_scaling, kwargs, expects):
    if expects == "sparse":
        init = random_sparse(
            dist="uniform", sr=sr, input_scaling=input_scaling, seed=42, **kwargs
        )
        w0 = init(*shape)
        w1 = random_sparse(
            *shape,
            dist="uniform",
            sr=sr,
            input_scaling=input_scaling,
            seed=42,
            **kwargs,
        )

        assert_allclose(w1.toarray(), w0.toarray(), atol=1e-12)

    if expects == "dense":
        init = random_sparse(
            dist="uniform", sr=sr, input_scaling=input_scaling, seed=42, **kwargs
        )
        w0 = init(*shape)
        w1 = random_sparse(
            *shape,
            dist="uniform",
            sr=sr,
            input_scaling=input_scaling,
            seed=42,
            **kwargs,
        )
        assert_allclose(w1, w0, atol=1e-12)

    if expects == "raise":
        with pytest.raises(Exception):
            init = random_sparse(
                dist="uniform", sr=sr, input_scaling=input_scaling, seed=42, **kwargs
            )
            w0 = init(*shape)
            w1 = random_sparse(
                *shape,
                dist="uniform",
                sr=sr,
                input_scaling=input_scaling,
                seed=42,
                **kwargs,
            )


@pytest.mark.parametrize(
    "shape,dtype,sparsity_type,kwargs,expects",
    [
        ((50, 50), np.float64, "csr", {"dist": "norm", "connectivity": 0.1}, "sparse"),
        ((50, 50), np.float32, "csc", {"dist": "norm", "connectivity": 0.1}, "sparse"),
        ((50, 50), np.int64, "coo", {"dist": "norm", "connectivity": 0.1}, "sparse"),
        ((50, 50), float, "dense", {"dist": "norm", "connectivity": 0.1}, "dense"),
    ],
)
def test_random_sparse_types(shape, dtype, sparsity_type, kwargs, expects):

    if expects == "sparse":
        init = random_sparse(
            dtype=dtype, sparsity_type=sparsity_type, seed=42, **kwargs
        )
        w0 = init(*shape)
        w1 = random_sparse(
            *shape, dtype=dtype, sparsity_type=sparsity_type, seed=42, **kwargs
        )

        assert_allclose(w1.toarray(), w0.toarray(), atol=1e-12)
        assert w0.dtype == dtype
        assert sparse.issparse(w0) and w0.format == sparsity_type

    if expects == "dense":
        init = random_sparse(
            dtype=dtype, sparsity_type=sparsity_type, seed=42, **kwargs
        )
        w0 = init(*shape)
        w1 = random_sparse(
            *shape, dtype=dtype, sparsity_type=sparsity_type, seed=42, **kwargs
        )

        assert_allclose(w1, w0, atol=1e-12)
        assert w0.dtype == dtype


@pytest.mark.parametrize(
    "initializer,shape,kwargs,expects",
    [
        (uniform, (50, 50), {"connectivity": 0.1}, "sparse"),
        (uniform, (50, 50, 50), {"connectivity": 0.1}, "dense"),
        (uniform, (50, 50), {"connectivity": 0.1, "sparsity_type": "dense"}, "dense"),
        (uniform, (50, 50), {"connectivity": 0.1, "high": 5.0, "low": 2.0}, "sparse"),
        (uniform, (50, 50), {"connectivity": 0.1, "high": 5.0, "low": "a"}, "raise"),
        (normal, (50, 50), {"connectivity": 0.1}, "sparse"),
        (normal, (50, 50, 50), {"connectivity": 0.1}, "dense"),
        (normal, (50, 50), {"connectivity": 0.1, "sparsity_type": "dense"}, "dense"),
        (normal, (50, 50), {"connectivity": 0.1, "loc": 5.0, "scale": 2.0}, "sparse"),
        (normal, (50, 50), {"connectivity": 0.1, "loc": 5.0, "scale": "a"}, "raise"),
        (bernoulli, (50, 50), {"connectivity": 0.1}, "sparse"),
        (bernoulli, (50, 50, 50), {"connectivity": 0.1}, "dense"),
        (bernoulli, (50, 50), {"connectivity": 0.1, "sparsity_type": "dense"}, "dense"),
        (bernoulli, (50, 50), {"connectivity": 0.1, "p": 0.9}, "sparse"),
        (bernoulli, (50, 50), {"connectivity": 0.1, "p": 5.0}, "raise"),
    ],
)
def test_dists(initializer, shape, kwargs, expects):
    if expects == "sparse":
        init = initializer(seed=42, **kwargs)
        w0 = init(*shape)
        w1 = initializer(*shape, seed=42, **kwargs)

        assert_allclose(w1.toarray(), w0.toarray(), atol=1e-12)

    if expects == "dense":
        init = initializer(seed=42, **kwargs)
        w0 = init(*shape)
        w1 = initializer(*shape, seed=42, **kwargs)
        assert_allclose(w1, w0, atol=1e-12)

    if expects == "raise":
        with pytest.raises(Exception):
            init = initializer(seed=42, **kwargs)
            w0 = init(*shape)
            w1 = initializer(*shape, seed=42, **kwargs)


def test_ones():
    w = ones(50, 50)
    assert_allclose(w, 1.0)

    w = ones(50, 50, dtype=np.float32)
    assert_allclose(w, 1.0)
    assert w.dtype == np.float32


def test_zeros():
    w = zeros(50, 50)
    assert_allclose(w, 0.0)

    w = zeros(50, 50, dtype=np.float32)
    assert_allclose(w, 0.0)
    assert w.dtype == np.float32

    with pytest.raises(ValueError):
        w = zeros(50, 50, sr=2.0)


@pytest.mark.parametrize(
    "N,expected", [(100, (100, 100)), (-1, Exception), ("foo", Exception)]
)
def test_fast_spectral_shape(N, expected):
    if expected is Exception:
        with pytest.raises(expected):
            fast_spectral_initialization(N)
    else:
        W = fast_spectral_initialization(N)
        assert W.shape == expected


@pytest.mark.parametrize("sr,proba", [(0.5, 0.1), (10.0, 0.5), (1.0, 1.0)])
def test_fast_spectral_features(sr, proba):
    W = fast_spectral_initialization(1000, sr=sr, connectivity=proba, seed=1234)

    if sparse.issparse(W):
        rho = max(
            abs(
                sparse.linalg.eigs(
                    W,
                    k=1,
                    which="LM",
                    maxiter=20 * W.shape[0],
                    return_eigenvectors=False,
                )
            )
        )
    else:
        rho = max(abs(linalg.eig(W)[0]))

    if proba == 0.0:
        assert_allclose(rho, 0.0)
    else:
        assert_allclose(rho, sr, rtol=1e-1)

    if 1.0 - proba < 1e-5:
        assert not sparse.issparse(W)
    if sparse.issparse(W):
        assert_allclose(
            np.count_nonzero(W.toarray()) / W.toarray().size, proba, rtol=1e-1
        )
    else:
        assert_allclose(np.count_nonzero(W) / W.size, proba, rtol=1e-1)


@pytest.mark.parametrize("sr,proba", [(1, -0.5), (1, 12), ("foo", 0.1)])
def test_fast_spectral_features_exception(sr, proba):
    with pytest.raises(Exception):
        fast_spectral_initialization(100, sr=sr, connectivity=proba)

    with pytest.raises(ValueError):
        fast_spectral_initialization(100, input_scaling=10.0, connectivity=proba)


def test_reproducibility_fsi():
    seed0 = default_rng(78946312)
    W0 = fast_spectral_initialization(
        100, sr=1.2, connectivity=0.4, seed=seed0
    ).toarray()

    seed1 = default_rng(78946312)
    W1 = fast_spectral_initialization(
        100, sr=1.2, connectivity=0.4, seed=seed1
    ).toarray()

    seed2 = default_rng(6135435)
    W2 = fast_spectral_initialization(
        100, sr=1.2, connectivity=0.4, seed=seed2
    ).toarray()

    assert_array_almost_equal(W0, W1)
    assert_raises(AssertionError, assert_array_almost_equal, W0, W2)


def test_sanity_checks():
    with pytest.raises(ValueError):
        _ = uniform(20, 20, degree=10, direction="all")
    with pytest.raises(ValueError):
        _ = uniform(30, degree=5, direction="in")
    with pytest.raises(ValueError):
        _ = uniform(30, 100, 10, degree=5, direction="in")
    with pytest.raises(ValueError):
        _ = bernoulli(30, 100, p=1.1)
    with pytest.raises(ValueError):
        _ = uniform(30, 100, low=1, high=0)


def test_ring_matrix():
    _ = ring(10, 10, weights=np.arange(1.0, 11.0), sr=1.0)
    W = ring(10, 10, input_scaling=2.0)

    assert W[1, 0] == 2.0 and W[0, -1] == 2.0

    # 1 on the 1st neuron, 0 elsewhere
    x0 = np.zeros((10, 1))
    x0[0, 0] = 1.0
    x = x0
    # loop all over the ring
    for i in range(10):
        x = W @ x

    assert np.all(x == 2**10 * x0)

    W_dense = ring(10, 10, input_scaling=2.0, sparsity_type="dense")
    assert_array_equal(W_dense, W.toarray())

    with pytest.raises(ValueError):
        _ = ring(10, 2, seed=1)


def test_line_matrix():
    _ = line(10, 10, weights=np.arange(1.0, 10.0), sr=1.0)
    W = line(10, 10, input_scaling=2.0)

    assert W[1, 0] == 2.0 and W[0, -1] == 0.0

    # 1 on the 1st neuron, 0 elsewhere
    x0 = np.zeros((10, 1))
    x0[0, 0] = 1.0
    x = x0
    # loop all over the line
    for i in range(10):
        x = W @ x

    assert np.all(x == 0.0)

    W_dense = line(10, 10, input_scaling=2.0, sparsity_type="dense")
    assert_array_equal(W_dense, W.toarray())

    with pytest.raises(ValueError):
        _ = line(10, 2, seed=1)


def test_orthogonal_matrix():
    W1 = orthogonal(10, 10, seed=1)
    W2 = orthogonal(10, 10, seed=1)

    assert np.all(np.isclose(W1, W2))

    assert np.all(np.isclose(W1 @ W1.T, np.eye(10)))

    with pytest.raises(ValueError):
        _ = orthogonal(10, 2, seed=1)

    with pytest.raises(ValueError):
        _ = orthogonal(10, 10, 10, seed=1)


def test_cluster_matrix():

    shape = 1000
    c = 10
    p_in = 0.1
    p_out = 0.01

    W1 = cluster(
        shape, shape, cluster=c, seed=1, p_in=p_in, p_out=p_out, distribution="normal"
    )
    W1 = W1.toarray()

    W2 = cluster(
        shape, shape, cluster=c, seed=1, p_in=p_in, p_out=p_out, distribution="normal"
    )
    W2 = W2.toarray()

    n_c = int(shape / c)

    # Check shape
    assert W1.shape == (shape, shape)

    # Assert that 2 matrices with same seed are equal
    assert_array_equal(W1, W2)

    # Check for cluster concentration
    def diagonal_concentration(band, w):
        n = w.shape[0]
        total = np.sum(np.abs(w))
        band_mask = np.abs(np.subtract.outer(np.arange(n), np.arange(n))) <= band
        diag_sum = np.sum(np.abs(w[band_mask]))
        diag_concentration = diag_sum / total
        # diagonal concentration ratio
        return diag_concentration

    W1_diag = diagonal_concentration(n_c, W1)
    np.random.seed(42)
    np.random.shuffle(W2)
    W2_diag = diagonal_concentration(n_c, W2)

    assert W1_diag > W2_diag

    # Connectivity test

    # p_in & p_out are 0 --> matrix is full of 0
    w_min = cluster(shape, shape, cluster=c, seed=1, p_in=0, p_out=0)
    w_min = w_min.toarray()
    assert np.all(w_min == 0)

    # p_in & p_out are 1 --> no 0s in the matrix
    w_max = cluster(shape, shape, cluster=c, seed=1, p_in=1, p_out=1)
    assert np.all(w_max != 0)

    # Check cluster density
    w_cluster_check = cluster(shape, shape, cluster=c, seed=1, p_in=p_in, p_out=p_out)
    w_cluster_check = w_cluster_check.toarray()

    non_zeros_in_clusters = 0
    for i in range(0, c):
        current_cluster = w_cluster_check[
            i * n_c : i * n_c + n_c, i * n_c : i * n_c + n_c
        ]
        non_zeros_in_clusters += np.sum(current_cluster != 0)
    assert non_zeros_in_clusters == (p_in * n_c * n_c * c)

    # Check for incorrect distribution
    with pytest.raises(ValueError):
        _ = cluster(
            shape,
            shape,
            cluster=c,
            seed=1,
            p_in=1,
            p_out=1,
            distribution="not_a_distribution",
        )

    # Check for invalid cluster and shape size
    with pytest.raises(ValueError):
        _ = cluster(
            shape, shape, cluster=3, seed=1, p_in=1, p_out=1, distribution="normal"
        )


def test_small_world_matrix():
    W1 = small_world(10, 10, seed=1)
    W2 = small_world(10, 10, seed=1)

    assert np.all(np.isclose(W1.toarray(), W2.toarray()))

    W1_big = small_world(1000, 1000, seed=1)
    W2_big = small_world(1000, 1000, seed=1)
    assert np.all(np.isclose(W1_big.toarray(), W2_big.toarray()))
    assert W1_big.shape == (1000, 1000)

    nb_close_neighbours = 2
    W3 = small_world(
        10,
        10,
        seed=1,
        sparsity_type="dense",
        nb_close_neighbours=nb_close_neighbours,
        proba_rewire=0.0,
    )

    assert np.all(np.diag(W3) == 0)

    assert np.all(np.sum(W3 != 0, axis=0) == nb_close_neighbours)
    assert np.all(np.sum(W3 != 0, axis=1) == nb_close_neighbours)

    W4 = small_world(10, 10, seed=1, nb_close_neighbours=0)

    assert np.all(np.sum(W4, axis=0) == 0)
    assert np.all(np.sum(W4, axis=1) == 0)

    with pytest.raises(ValueError):
        _ = small_world(10, 2, seed=1)

    with pytest.raises(ValueError):
        _ = small_world(10, 10, 10, seed=1)
    with pytest.raises(ValueError):
        _ = small_world(10, 10, seed=1, nb_close_neighbours=3)
    with pytest.raises(ValueError):
        _ = small_world(10, 10, seed=1, proba_rewire=1.5)
    with pytest.raises(ValueError):
        _ = small_world(10, 10, seed=1, proba_rewire=-0.5)
    with pytest.raises(ValueError):
        _ = small_world(10, 10, seed=1, distribution="not_a_distribution")


def test_small_world_sparsity_type():
    W1 = small_world(100, 100, seed=1, sparsity_type="csr")
    assert isinstance(W1, sparse.sparray)
    assert W1.format == "csr"



from typing import Callable, Literal, Optional, Sequence, Union

import numpy as np
import scipy.sparse as sp
from numpy.random import Generator

from reservoirpy.utils.data_validation import check_node_input

from ..activationsfunc import get_function, tanh
from ..mat_gen import bernoulli, uniform
from ..node import TrainableNode
from ..type import (
    NodeInput,
    State,
    Timeseries,
    Timestep,
    Weights,
    is_array,
    is_multiseries,
)
from ..utils.random import rand_generator


class LocalPlasticityReservoir(TrainableNode):
    """
    A reservoir that learns its recurrent weights W through a local
    learning rule selected by the ``learning_rule`` hyperparameter.

    Reservoir states are updated with the external equation:

    .. math::

        & r[t+1] = (1 - lr)*r[t] + lr*(W r[t] + W_{in} u[t+1] + bias) \\\\
        & x[t+1] = f(r[t+1])

    Where :math:`f` is the activation function.

    Then the local rule is applied each timestep to update W.

    .. math::

        W_{ij} \\leftarrow W_{ij} + \\Delta W_{ij}

    Supported rules:
      ``oja``:
        :math:`\\Delta W_{ij} = \\eta y (x - y W_{ij})`
      ``anti-oja`` [1]_ [2]_ [3]_ :
        :math:`\\Delta W_{ij} = - \\eta y (x - y W_{ij})`
      ``hebbian`` [4]_ :
        :math:`\\Delta W_{ij} = \\eta x y`
      ``anti-hebbian``:
        :math:`\\Delta W_{ij} = - \\eta x y`
      ``bcm`` [2]_ :
        :math:`\\Delta W_{ij} = \\eta x y (y - \\theta_{BCM})`

    Where :math:`x` represents the pre-synaptic state and :math:`y` represents
    the post-synaptic state of the neuron.

    For "`bcm`", you can set a threshold ``bcm_theta`` (default `0.0`).

    If ``synapse_normalization=True``, then after each local-rule update
    on a row i of W, the row is rescaled to unit L2 norm. [4]_



    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    local_rule : str, default to `oja`
        One of `"oja"`, `"anti-oja"`, `"hebbian"`, `"anti-hebbian"`, `"bcm"`.
    eta : float, default to 1e-3.
        Local learning rate for the weight update.
    bcm_theta : float, default to 0.0
        The threshold used in the "bcm" rule.
    synapse_normalization : bool, default to True
        If True, L2-normalize each row of W after its update.
    epochs : int, default to 1
        Number of training iterations.
    sr : float, default to 1.0
        Spectral radius of recurrent weight matrix.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is
        used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.uniform`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    activation : {"tanh", "sigmoid"}, default to "tanh"
        Reservoir units activation function.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    name : str, optional
        Node name.


    References
    ----------

    .. [1] Babinec, Š., & Pospíchal, J. (2007). Improving the prediction
           accuracy of echo state neural networks by anti-Oja’s learning.
           In International Conference on Artificial Neural Networks (pp. 19-28).
           Berlin, Heidelberg: Springer Berlin Heidelberg.
           https://doi.org/10.1007/978-3-540-74690-4_3

    .. [2] Yusoff, M. H., Chrol-Cannon, J., & Jin, Y. (2016). Modeling neural
           plasticity in echo state networks for classification and regression.
           Information Sciences, 364, 184-196.
           https://doi.org/10.1016/j.ins.2015.11.017

    .. [3] Morales, G. B., Mirasso, C. R., & Soriano, M. C. (2021). Unveiling
           the role of plasticity rules in reservoir computing. Neurocomputing,
           461, 705-715. https://doi.org/10.1016/j.neucom.2020.05.127

    .. [4] Wang, X., Jin, Y., & Hao, K. (2021). Synergies between synaptic and
           intrinsic plasticity in echo state networks. Neurocomputing,
           432, 32-43. https://doi.org/10.1016/j.neucom.2020.12.007

    Example
    -------
    >>> reservoir = LocalPlasticityReservoir(
    ...     units=100, sr=0.9, local_rule="hebbian",
    ...     eta=1e-3, epochs=5, synapse_normalization=True
    ... )
    >>> # Fit on data timeseries
    >>> reservoir.fit(X_data, warmup=10)
    >>> # Then run
    >>> states = reservoir.run(X_data)
    """

    #: Number of neuronal units in the reservoir.
    units: int
    #: Local learning rate for the weight update.
    eta: float
    #: The threshold used in the "bcm" rule.
    bcm_theta: float
    #: If True, L2-normalize each row of W after its update.
    synapse_normalization: bool
    #: Number of training iterations.
    epochs: int
    #: Leaking rate (1.0 by default) (:math:`\mathrm{lr}`).
    lr: float
    #: Spectral radius of ``W`` (optional).
    sr: float
    #: Input scaling (float or array) (1.0 by default).
    input_scaling: Union[float, Sequence]
    #: Connectivity (or density) of ``Win`` (0.1 by default).
    input_connectivity: float
    #: Connectivity (or density) of ``Wfb`` (0.1 by default).
    rc_connectivity: float
    #: Input weights matrix (:math:`\mathbf{W}_{in}`).
    Win: Weights
    #: Recurrent weights matrix (:math:`\mathbf{W}`).
    W: Weights
    #: Bias vector (:math:`\mathbf{b}`).
    bias: Weights
    #: Activation of the reservoir units (tanh by default) (:math:`f`).
    activation: Callable
    #: Type of matrices elements. By default, ``np.float64``.
    dtype: type
    #: A random state generator. Used for generating Win and W.
    rng: Generator

    def __init__(
        self,
        units: Optional[int] = None,
        # local rule choice
        local_rule: Literal[
            "oja", "anti-oja", "hebbian", "anti-hebbian", "bcm"
        ] = "oja",
        eta: float = 1e-3,
        bcm_theta: float = 0.0,
        synapse_normalization: bool = False,
        epochs: int = 1,
        # standard reservoir params
        sr: float = 1.0,
        lr: float = 1.0,
        input_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: float = 0.1,
        rc_connectivity: float = 0.1,
        Win: Union[Weights, Callable] = bernoulli,
        W: Union[Weights, Callable] = uniform,
        bias: Union[Weights, Callable] = bernoulli,
        activation: Union[str, Callable] = tanh,
        input_dim: Optional[int] = None,
        seed: Optional[Union[int, np.random.Generator]] = None,
        dtype: type = np.float64,
        name: Optional[str] = None,
    ):
        self.units = units
        self.eta = eta
        self.bcm_theta = bcm_theta
        self.synapse_normalization = synapse_normalization
        self.epochs = epochs
        self.lr = lr
        self.sr = sr
        self.input_scaling = input_scaling
        self.input_connectivity = input_connectivity
        self.rc_connectivity = rc_connectivity
        self.Win = Win
        self.W = W
        self.bias = bias
        self.activation = get_function(activation)
        self.dtype = dtype
        self.rng = rand_generator(seed=seed)
        self.initialized = False
        self.name = name

        # set units / output_dim
        if units is None and not is_array(W):
            raise ValueError(
                "'units' parameter must not be None if 'W' parameter is not a matrix."
            )
        if units is not None and is_array(W) and W.shape[-1] != units:
            raise ValueError(
                f"Both 'units' and 'W' are set but their dimensions doesn't match: "
                f"{units} != {W.shape[-1]}."
            )
        self.units = units if units is not None else W.shape[-1]
        self.output_dim = self.units

        # set input_dim (if possible)
        if input_dim is not None and is_array(Win) and Win.shape[-1] != input_dim:
            raise ValueError(
                f"Both 'input_dim' and 'Win' are set but their dimensions doesn't "
                f"match: {input_dim} != {Win.shape[-1]}."
            )
        self.input_dim = Win.shape[-1] if is_array(Win) else input_dim

        # Validate local rule name
        local_rule = local_rule.lower()

        def oja(weights, pre_state, post_state):
            return eta * post_state * (pre_state - post_state * weights)

        def anti_oja(weights, pre_state, post_state):
            return -eta * post_state * (pre_state - post_state * weights)

        def hebbian(weights, pre_state, post_state):
            return eta * post_state * pre_state

        def anti_hebbian(weights, pre_state, post_state):
            return -eta * post_state * pre_state

        def bcm(weights, pre_state, post_state):
            return eta * post_state * (post_state - bcm_theta) * pre_state

        rules = {
            "oja": oja,
            "anti-oja": anti_oja,
            "hebbian": hebbian,
            "anti-hebbian": anti_hebbian,
            "bcm": bcm,
        }
        if not local_rule in rules:
            raise ValueError(
                f"Unknown learning rule '{local_rule}'. Choose from: "
                "['oja', 'anti-oja', 'hebbian', 'anti-hebbian', 'bcm']."
            )

        self.increment = rules[local_rule]

    def initialize(self, x: Optional[Union[NodeInput, Timestep]]):

        # set input_dim
        self._set_input_dim(x)

        [Win_rng, W_rng, bias_rng] = self.rng.spawn(3)

        if callable(self.Win):
            self.Win = self.Win(
                self.units,
                self.input_dim,
                input_scaling=self.input_scaling,
                connectivity=self.input_connectivity,
                dtype=self.dtype,
                seed=Win_rng,
            )

        if callable(self.W):
            self.W = self.W(
                self.units,
                self.units,
                sr=self.sr,
                connectivity=self.rc_connectivity,
                dtype=self.dtype,
                seed=W_rng,
            )

        if callable(self.bias):
            self.bias = self.bias(
                self.units,
                connectivity=1.0,
                dtype=self.dtype,
                seed=bias_rng,
            )

        self.state = {
            "out": np.zeros((self.units,)),
            "internal": np.zeros((self.units,)),
        }

        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        W = self.W  # NxN
        Win = self.Win  # NxI
        bias = self.bias  # N or float
        f = self.activation
        lr = self.lr
        s = state["out"]

        next_state = W @ s + Win @ x + bias
        next_state = (1 - lr) * s + lr * next_state

        return {"internal": next_state, "out": f(next_state)}

    def fit(
        self, x: NodeInput, y: None = None, warmup: int = 0
    ) -> "LocalPlasticityReservoir":
        check_node_input(x, expected_dim=self.input_dim)

        if not self.initialized:
            self.initialize(x)

        increment = self.increment
        do_norm = self.synapse_normalization

        def _local_synaptic_plasticity(seq: Timeseries):
            """
            Apply the local learning rule (Oja, Anti-Oja, Hebbian, Anti-Hebbian, BCM)
            to update the recurrent weight matrix W.

            If `synapse_normalization=True`, then each row of W is L2-normalized
            immediately after the local rule update.
            """
            for u in seq:
                pre_state = self.state["internal"]  # (units,)
                post_state = self._step(self.state, u)["internal"]  # (units,)
                # Vectorized update of nonzero elements based on the chosen rule.
                (rows, cols, data) = sp.find(self.W)
                self.W[rows, cols] += increment(data, pre_state[cols], post_state[rows])
                # Optionally normalize each row.
                if do_norm:
                    # Compute the L2 norm per row for the updated data.
                    row_norms = np.sqrt(np.sum(self.W**2, axis=1)).reshape(-1, 1)
                    safe_norms = np.where(row_norms > 0, row_norms, 1)
                    self.W[:] /= safe_norms[:]

        for _epoch in range(self.epochs):
            if is_multiseries(x):
                for seq in x:
                    _local_synaptic_plasticity(seq[warmup:])
            else:
                _local_synaptic_plasticity(x[warmup:])
        return self



import os

from setuptools import find_packages, setup

NAME = "reservoirpy"

__version__ = ""
version_file = os.path.join("reservoirpy", "_version.py")
with open(version_file) as f:
    exec(f.read())

AUTHOR = "Xavier Hinaut"
AUTHOR_EMAIL = "xavier.hinaut@inria.fr"

MAINTAINERS = "Xavier Hinaut, Paul Bernard"
MAINTAINERS_EMAIL = "xavier.hinaut@inria.fr, paul.bernard@inria.fr"

DESCRIPTION = "A simple and flexible code for Reservoir " "Computing architectures like Echo State Networks."

with open("README.md", "r", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

URL = "https://github.com/reservoirpy/reservoirpy"
DOWNLOAD_URL = f"{URL}/v{__version__}.tar.gz"

INSTALL_REQUIRES = [
    "joblib>=0.14.1",
    "numpy>=1.21.1",
    "scipy>=1.4.1",
]

EXTRA_REQUIRES = {
    "hyper": ["hyperopt", "matplotlib>=2.2.0", "tqdm>=4.43.0"],
    "sklearn": ["scikit-learn>=0.24.2"],
}

PROJECT_URLS = {
    "Bug Tracker": "https://github.com/reservoirpy/reservoirpy/issues",
    "Documentation": "https://reservoirpy.readthedocs.io/en/latest/index.html",
    "Source Code": URL,
    "Release notes": "https://github.com/reservoirpy/reservoirpy/releases",
}

if __name__ == "__main__":
    setup(
        name=NAME,
        version=__version__,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINERS,
        maintainer_email=MAINTAINERS_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        url=URL,
        project_urls=PROJECT_URLS,
        download_url=DOWNLOAD_URL,
        packages=find_packages(),
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: Implementation :: PyPy",
        ],
        python_requires=">=3.9",
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRA_REQUIRES,
        package_data={"reservoirpy": ["datasets/santafe_laser.npy"]},
    )



from typing import Optional

from reservoirpy.node import Node, TrainableNode
from reservoirpy.nodes.io import Input
from reservoirpy.nodes.reservoir import Reservoir
from reservoirpy.nodes.ridge import Ridge
from reservoirpy.type import Edge

from .model import Model
from .utils.model_utils import obj_from_kwargs


class ESN(Model):
    """Simple Echo State Network.

    This class is provided as a wrapper for a simple reservoir connected to a
    readout.

    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    reservoir : Node, optional
        A Node instance to use as a reservoir,
        such as a :py:class:`~reservoirpy.nodes.Reservoir` node.
    readout : Node, optional
        A Node instance to use as a readout,
        such as a :py:class:`~reservoirpy.nodes.Ridge` node
        (only this one is supported).
    feedback : bool, defaults to False
        If True, the readout is connected to the reservoir through
        a feedback connection.
    input_to_readout : bool, defaults to False
        If True, the input is directly fed to the readout. See
        :ref:`/user_guide/advanced_demo.ipynb#Input-to-readout-connections`.
    **kwargs
        Arguments passed to the reservoir and readout.

    See Also
    --------
    Reservoir
    Ridge

    Example
    -------
    >>> from reservoirpy import ESN
    >>>
    >>> model = ESN(units=100, sr=0.9, ridge=1e-6)
    >>>
    """

    #: A :py:class:`~reservoirpy.nodes.Reservoir` or a :py:class:`~reservoirpy.nodes.NVAR` instance.
    reservoir: Node
    #: A :py:class:`~reservoirpy.nodes.Ridge` instance.
    readout: TrainableNode
    #: Is readout connected to reservoir through feedback (False by default).
    feedback: bool
    #: Does the readout directly receives the input (False by default).
    input_to_readout: bool
    #: Input node, if ``input_to_readout`` is set to True
    input_node: Optional[Input] = None

    def __init__(
        self,
        reservoir: Optional[Reservoir] = None,
        readout: Optional[Ridge] = None,
        feedback=False,
        input_to_readout=False,
        **kwargs,
    ):
        if reservoir is None:
            reservoir = obj_from_kwargs(Reservoir, kwargs)

        if readout is None:
            # avoid argument name collision
            if "bias" in kwargs:
                kwargs.pop("bias")
            if "readout_bias" in kwargs:
                kwargs["bias"] = kwargs["readout_bias"]
                kwargs.pop("readout_bias")
            if "input_dim" in kwargs:
                kwargs.pop("input_dim")
            readout = obj_from_kwargs(Ridge, kwargs)

        nodes: list[Node] = [reservoir, readout]
        edges: list[Edge] = [(reservoir, 0, readout)]

        if feedback:
            edges.append((readout, 1, reservoir))

        if input_to_readout:
            input_node = Input()
            nodes.append(input_node)
            edges.append((input_node, 0, reservoir))
            edges.append((input_node, 0, readout))
            self.input_node = input_node

        self.reservoir = reservoir
        self.readout = readout
        self.feedback = feedback
        self.input_to_readout = input_to_readout

        super(ESN, self).__init__(nodes=nodes, edges=edges)



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from reservoirpy.utils.model_utils import data_from_buffer, join_data


def test_join_data():
    x1 = np.zeros((10, 2))
    out = join_data(x1)
    assert_array_equal(x1, out)

    x2 = np.ones((10, 12))
    out = join_data(x1, x2)
    assert out.shape == (10, 2 + 12)

    list1 = [np.ones((2 * i + 5, 7)) for i in range(3)]
    list2 = [np.ones((2 * i + 5, 3)) for i in range(3)]
    out = join_data(list1, list2, list1)
    assert isinstance(out, list)
    for el in out:
        assert el.shape[-1] == 7 + 3 + 7


def test_data_from_buffer():
    # simple timeseries
    x = np.arange(10).reshape(-1, 1)
    buffer = np.arange(-5, 0).reshape(-1, 1)
    new_buffer, new_x = data_from_buffer(buffer, x)
    assert new_buffer.shape == buffer.shape
    assert new_x.shape == x.shape
    np.testing.assert_array_equal(new_buffer, x[-5:])

    # multiseries 3D array
    x = np.arange(20).reshape(2, -1, 1)
    buffer = np.arange(-5, 0).reshape(-1, 1)
    new_buffer, new_x = data_from_buffer(buffer, x)
    assert new_buffer.shape == buffer.shape
    assert new_x.shape == x.shape
    np.testing.assert_array_equal(new_buffer, x[-1, -5:])

    # list of timeseries
    x = [np.arange(100 * i, 100 * i + 10).reshape(-1, 1) for i in range(3)]
    buffer = np.arange(-5, 0).reshape(-1, 1)
    new_buffer, new_x = data_from_buffer(buffer, x)
    assert len(new_x) == len(x)
    np.testing.assert_array_equal(new_buffer, x[-1][-5:])



# Author: Nathan Trouvain at 07/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import sys
from pathlib import Path
from typing import List, Union

import numpy as np

DATA_FOLDER = Path.home() / Path("reservoirpy-data")


def _get_data_folder(folder_path=None):
    if folder_path is None:
        folder_path = DATA_FOLDER
    else:
        folder_path = Path(folder_path)

    if not folder_path.exists():
        folder_path.mkdir(parents=True)

    return folder_path


def one_hot_encode(y: Union[np.ndarray, List]):
    """Encode categorical features as a one-hot numeric array.

    This functions creates a trailing column for each class from the dataset. This function also supports inputs as
    lists of numpy arrays to stay compatible with the ReservoirPy format.

    Accepted inputs and corresponding outputs:

    - array of shape (n, ) or (n, 1) or list of length n -> array of shape (n, n_classes)
    - array of shape (n, m) or (n, m, 1) -> array of shape (n, m, n_classes)
    - list of arrays with shape (m, ) or (m, 1) -> list of arrays with shape (n, n_classes)

    Parameters
    ----------
    X: array or list of categorical values, or list of array of categorical values
        The data to determine the categories of each features.

    Returns
    -------
    array or list. See above for details.
        One-hot encoded dataset

    Example
    -------
    >>> from reservoirpy.datasets import one_hot_encode
    >>> X = np.random.normal(size=(10, 100, 1))  # 10 series, 100 timesteps
    >>> y = np.mean(X, axis=(1,2)) > 0. # a boolean for each series
    >>> print(y)
    [ True False False False  True False  True  True  True False]
    >>> y_encoded, classes = one_hot_encode(y)
    >>> y_encoded
    array([ [0., 1.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [0., 1.],
            [1., 0.],
            [0., 1.],
            [0., 1.],
            [0., 1.],
            [1., 0.]])
    >>> classes
    array([False,  True])

    """
    if isinstance(y, list) and isinstance(y[0], np.ndarray):  # multi-sequence
        # treat it as one long timeseries before re-separating them
        series_lengths = [series.shape[0] for series in y]
        series_split_indices = np.cumsum(series_lengths)[:-1]
        concatenated_series = np.concatenate(y)
        concatenated_encoded, classes = one_hot_encode(concatenated_series)
        encoded = np.split(concatenated_encoded, series_split_indices)
        return encoded, classes

    y = np.array(y)

    if y.shape[-1] == 1:
        y = y.reshape(y.shape[:-1])

    classes, y_class_indices = np.unique(y, return_inverse=True)
    y_class_indices = y_class_indices.reshape(y.shape)
    nb_classes = len(classes)
    encoder = np.eye(nb_classes)
    y_encoded = encoder[y_class_indices]
    return y_encoded, classes


def from_aeon_classification(
    X: Union[np.ndarray, List[np.ndarray]],
):
    """Converts a dataset in the `Aeon <https://aeon-toolkit.org/>`_ classification format into a ReservoirPy-compatible format.

    The Aeon library provides many classical classification datasets, notably all the benchmark datasets from the
    `<https://timeseriesclassification.com>`_ website. You can also use aeon to load datasets in various fileformats.

    Parameters
    ----------
    X : array-like of shape (n_timeseries, n_dimensions, n_timesteps) or list of arrays of shape (n_dimensions, n_timesteps)
        Input data in the aeon dataset format for classification

    Returns
    -------
    X : array of shape (n_timeseries, n_timesteps, n_dimensions) or list of arrays of shape (n_timesteps, n_dimensions)
        Input data in the ReservoirPy dataset format


    Examples
    --------

    >>> from aeon.datasets import load_classification
    >>> X, y = load_classification("FordA")
    >>> print(X.shape)
    (4921, 1, 500)
    >>> from reservoirpy.datasets import from_aeon_classification
    >>> X_ = from_aeon_classification(X)
    >>> print(X_.shape)
    (4921, 500, 1)
    """
    X_out: np.ndarray | List[np.ndarray]

    if isinstance(X, list):
        X_out = [np.swapaxes(np.array(series), 0, 1) for series in X]
        return X_out

    if not isinstance(X, np.ndarray):
        if np.array(X).shape == ():
            raise TypeError(f"X must be numpy array-like or a list, got {type(X)}")
        X = np.array(X)

    if not len(X.shape) == 3:
        raise ValueError(
            f"Expected a 3-dimensional array, got {len(X.shape)} dimensions."
        )

    X_out = np.swapaxes(X, 1, 2)
    return X_out



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_almost_equal

from reservoirpy.activationsfunc import identity, relu, sigmoid, softmax, softplus, tanh


@pytest.mark.parametrize(
    "value, expected",
    [
        (np.array([1, 2, 3]), np.exp([1, 2, 3]) / np.sum(np.exp([1, 2, 3]))),
        (np.array(1), np.exp(1) / np.sum(np.exp(1))),
        (np.array([0, 0]), np.array([0.5, 0.5])),
    ],
)
def test_softmax(value, expected):
    result = softmax(value)

    assert_almost_equal(np.sum(result), 1.0)
    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        (np.array(0), np.log(1 + np.exp(0))),
        (np.array([0, 1, 2]), np.log(1 + np.exp([0, 1, 2]))),
        (np.array([-2, -1]), np.log(1 + np.exp([-2, -1]))),
    ],
)
def test_softplus(value, expected):
    result = softplus(value)

    assert np.any(result > 0.0)
    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "value",
    [
        (np.array([1, 2, 3])),
        (np.array([1])),
        (np.array(0)),
        (np.array([0.213565165, 0.1, 1.064598495615132])),
    ],
)
def test_identity(value):
    result = identity(value)
    val = np.asanyarray(value)

    assert np.any(result == val)


@pytest.mark.parametrize(
    "value, expected", [(np.array([1, 2, 3]), np.tanh([1, 2, 3])), (0, np.tanh(0))]
)
def test_tanh(value, expected):
    result = tanh(value)

    assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "value, expected",
    [
        (np.array([1, 2, 3]), 1 / (1 + np.exp(-np.array([1, 2, 3])))),
        (np.array(0), 1 / (1 + np.exp(0))),
        (np.array([-1000, -2]), np.array([0.0, 1.35e-1])),
    ],
)
def test_sigmoid(value, expected):
    result = sigmoid(value)
    assert_array_almost_equal(result, expected, decimal=1)


@pytest.mark.parametrize(
    "value, expected",
    [
        (np.array([1, 2, 3]), np.array([1, 2, 3])),
        (np.array([-1, -10, 5]), np.array([0, 0, 5])),
        (np.array(0), np.array(0)),
        (np.array([[1, 2, 3], [-1, 2, 3]]), np.array([[1, 2, 3], [0, 2, 3]])),
    ],
)
def test_relu(value, expected):
    result = relu(value)
    assert_array_almost_equal(result, expected)



# Author: Nathan Trouvain at 14/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import itertools as it
from math import comb
from typing import Optional, Sequence, Union

import numpy as np

from ..node import Node
from ..type import NodeInput, State, Timestep


class NVAR(Node):
    """Non-linear Vector AutoRegressive machine.

    NVAR is implemented as described in [1]_.

    The state :math:`\\mathbb{O}_{total}` of the NVAR first contains a series of linear
    features :math:`\\mathbb{O}_{lin}` made of input data concatenated
    with delayed inputs:

    .. math::

        \\mathbb{O}_{lin}[t] = \\mathbf{X}[t] \\oplus \\mathbf{X}[t - s] \\oplus
        \\mathbf{X}[t - 2s] \\oplus \\dots \\oplus \\mathbf{X}[t - (k-1)s]

    where :math:`\\mathbf{X}[t]` are the inputs at time :math:`t`, :math:`k` is the
    delay and :math:`s` is the strides (only one input every :math:`s`
    inputs within the delayed inputs is used).
    The operator :math:`\\oplus` denotes the concatenation.

    In addition to these linear features, nonlinear representations
    :math:`\\mathbb{O}_{nonlin}^n` of the inputs are constructed using all unique
    monomials of order :math:`n` of these inputs:

    .. math::

        \\mathbb{O}_{nonlin}^n[t] = \\mathbb{O}_{lin}[t] \\otimes \\mathbb{O}_{lin}[t]
        \\overbrace{\\otimes \\dots \\otimes}^{n-1~\\mathrm{times}} \\mathbb{O}_{lin}[t]

    where :math:`\\otimes` is the operator denoting an outer product followed by the
    selection of all unique monomials generated by this outer product.

    Note
    ----

        Under the hood,
        this product is computed by finding all unique combinations
        of input features and multiplying each combination of terms.

    Finally, all representations are gathered to form the final feature
    vector :math:`\\mathbb{O}_{total}`:

    .. math::

        \\mathbb{O}_{total} = \\mathbb{O}_{lin}[t] \\oplus \\mathbb{O}_{nonlin}^n[t]


    Parameters
    ----------
    delay : int
        Maximum delay of inputs.
    order : int
        Order of the non-linear monomials.
    strides : int, default to 1
        Strides between delayed inputs.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.

    References
    ----------
    .. [1] Gauthier, D. J., Bollt, E., Griffith, A., & Barbosa, W. A. S. (2021).
           Next generation reservoir computing. Nature Communications, 12(1), 5564.
           https://doi.org/10.1038/s41467-021-25801-2

    Example
    -------

    >>> import numpy as np
    >>> from reservoirpy.nodes import NVAR, Ridge
    >>> nvar = NVAR(delay=2, order=2, strides=1)
    >>> readout = Ridge(ridge=2.5e-6, output_dim=3)
    >>> model = nvar >> readout

    Using the :py:func:`~reservoirpy.datasets.lorenz` timeseries and learning to
    predict the next difference:

    >>> from reservoirpy.datasets import lorenz
    >>> X = lorenz(5400, x0=[17.677, 12.931, 43.914], h=0.025, method="RK23")
    >>> Xi  = X[:600]
    >>> dXi = X[1:601] - X[:600]  # difference u[t+1] - u[t]
    >>> Y_test = X[600:]  # testing data
    >>> _ = model.fit(Xi, dXi, warmup=200)

    We can now predict the differences and integrate these predictions:

    >>> u = X[600]
    >>> res = np.zeros((5400-600, readout.output_dim))
    >>> for i in range(5400-600):
    ...     u = u + model(u)
    ...     res[i, :] = u
    ...

    .. plot:: ./api/nvar_example.py

    """

    #: Time window over the inputs (of shape (delay * strides, features)).
    store: np.ndarray
    #: Maximum delay of inputs (:math:`k`).
    delay: int
    #: Order of the non-linear monomials (:math:`n`).
    order: int
    #: Strides between delayed inputs, by default 1 (:math:`s`).
    strides: int

    def __init__(
        self,
        delay: int,
        order: int,
        strides: int = 1,
        input_dim: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.store = None
        self._monomial_idx = None
        self.delay = delay
        self.order = order
        self.strides = strides
        self.name = name
        self.initialized = False
        self.state = {}
        self.input_dim = input_dim
        self.output_dim = None

    def initialize(self, x: Union[NodeInput, Timestep]):

        # set input_dim
        self._set_input_dim(x)

        order = self.order
        delay = self.delay
        strides = self.strides

        linear_dim = delay * self.input_dim
        # number of non linear components is (d + n - 1)! / (d - 1)! n!
        # i.e. number of all unique monomials of order n made from the
        # linear components.
        nonlinear_dim = comb(linear_dim + order - 1, order)

        output_dim = int(linear_dim + nonlinear_dim)

        self.output_dim = output_dim

        # for each monomial created in the non linear part, indices
        # of the n components involved, n being the order of the
        # monomials. Pre-compute them to improve efficiency.
        idx = np.array(
            list(it.combinations_with_replacement(np.arange(linear_dim), order))
        )

        self._monomial_idx = idx

        # to store the k*s last inputs, k being the delay and s the strides
        self.store = np.zeros((delay * strides, self.input_dim))
        self.initialized = True

    def _step(self, state: tuple, x: Timestep) -> State:
        store = self.store
        strides = self.strides
        idxs = self._monomial_idx
        output_dim = self.output_dim

        # store the current input
        new_store = np.roll(store, 1, axis=0)
        new_store[0] = x
        self.store = new_store

        output = np.zeros((output_dim,))

        # select all previous inputs, including the current, with strides
        linear_feats = np.ravel(new_store[::strides, :])
        linear_len = linear_feats.shape[0]

        output[:linear_len] = linear_feats

        # select monomial terms and compute them
        output[linear_len:] = np.prod(linear_feats[idxs], axis=1)

        return {"out": output}
